/*
  create_cdb -- a program for creating random (possibly corrupt) CDB files
  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <strings.h>
#include <getopt.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include "spamdyke.h"
#include "cdb.h"

#define STRLEN(X)       ((int)(sizeof(X) - 1))

#define MAX_BUF         1023

#define ALPHABET        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.:/?;!@#$%^&*()"

struct record
  {
  uint32_t hash;
  int index;
  int slot_num;
  int strlen_key;
  char key[(MAX_BUF * 2) + 1];
  int strlen_data;
  char data[MAX_BUF + 1];
  int corrupt_overlength_key;
  int corrupt_hash_index_mismatch;
  int corrupt_hash_key_mismatch;
  int corrupt_index_offset_oversize;
  int corrupt_hash_offset_oversize;
  int corrupt_key_length_oversize;
  int corrupt_data_length_oversize;
  };

struct main_index_record
  {
  uint32_t *hash_values;
  uint32_t num_slots;
  int *slots;
  int corrupt_index_offset_oversize;
  };

void usage()
  {
  printf("USAGE: create_cdb [ -v ] -f CDB_FILE [ -n NUM_ENTRIES ] "
      "[ -x NUM_HASH_INDEX_MISMATCHES ] "
      "[ -i NUM_INDEX_OFFSET_OVERSIZES ] "
      "[ -h NUM_HASH_OFFSET_OVERSIZES ] "
      "[ -m NUM_HASH_KEY_MISMATCHES ] "
      "[ -k NUM_KEY_LENGTH_OVERSIZES ] "
      "[ -d NUM_DATA_LENGTH_OVERSIZES ] "
      "[ -o NUM_OVERLENGTH_KEYS ] "
    "\n\n"
    "Creates a CDB file filled with random data for testing purposes.\n"
    "\n"
    "-f CDB_FILE\n"
    "   Create the file with the given name.\n"
    "\n"
    "-n NUM_ENTRIES\n"
    "   Fill the file with NUM_ENTRIES randomly generated records. Default: 0 (create\n"
    "   a valid CDB file with no data)\n"
    "\n"
    "-v\n"
    "   Generate verbose output as the program runs.  Giving this flag multiple times\n"
    "   increases the verbosity.\n"
    "\n"
    "-x NUM_HASH_INDEX_MISMATCHES\n"
    "   Corrupt NUM_HASH_INDEX_MISMATCHES records by saving them under the wrong main\n"
    "   index number, given their hash value. Default: 0\n"
    "\n"
    "-i NUM_INDEX_OFFSET_OVERSIZES\n"
    "   Corrupt NUM_INDEX_OFFSET_OVERSIZES main index records by setting their offset\n"
    "   fields to extend beyond the end of the file. Default: 0\n"
    "\n"
    "-h NUM_HASH_OFFSET_OVERSIZES\n"
    "   Corrupt NUM_HASH_OFFSET_OVERSIZES records by setting their offset fields\n"
    "   within the hash index to a location beyond the end of the file. Default: 0\n"
    "\n"
    "-m NUM_HASH_KEY_MISMATCHES\n"
    "   Corrupt NUM_HASH_KEY_MISMATCHES records by saving them with a key that does\n"
    "   not match their hash value. Default: 0\n"
    "\n"
    "-k NUM_KEY_LENGTH_OVERSIZES\n"
    "   Corrupt NUM_KEY_LENGTH_OVERSIZES records by setting the length of their key\n"
    "   fields to extend beyond the end of the file. Default 0\n"
    "\n"
    "-d NUM_DATA_LENGTH_OVERSIZES\n"
    "   Corrupt NUM_DATA_LENGTH_OVERSIZES records by setting the length of their data\n"
    "   fields to extend beyond the end of the file. Default: 0\n"
    "\n"
    "-o NUM_OVERLENGTH_KEYS\n"
    "   Generate NUM_OVERLENGTH_KEYS record with very long keys (over " STRINGIFY(MAX_BUF) " bytes).\n"
    "   Default: 0\n"
    "\n"
    "NOTE: No record will be corrupted in multiple ways.  If the total number of\n"
    "requested corruptions is greater than NUM_ENTRIES, not all of them will be\n"
    "performed.\n"
  );

  exit(0);

  return;
  }

uint32_t htole(uint32_t target_int)
  {
  static int little_endian = -1;
  uint32_t return_value;
  int i;
  union
    {
    uint32_t integer;
    unsigned char string[4];
    } convert;

  if (little_endian == -1)
    {
    convert.string[0] = 0x01;
    convert.string[1] = 0x00;
    convert.string[2] = 0x00;
    convert.string[3] = 0x00;
    little_endian = (convert.integer == 1) ? 1 : 0;
    }

  if (!little_endian)
    {
    for (i = 0; i < 4; i++)
      convert.string[i] = (unsigned char)((target_int >> (i * 8)) & 0xFF);

    return_value = convert.integer;
    }
  else
    return_value = target_int;

  return(return_value);
  }

uint32_t cdb_hash(char *target_key, int target_strlen)
  {
  uint32_t return_value;
  int i;

  return_value = CDB_HASH_SEED;
  if (target_key != NULL)
    for (i = 0; i < target_strlen; i++)
      return_value = ((return_value << 5) + return_value) ^ target_key[i];

  return(return_value);
  }

int main(int argc, char *argv[])
  {
  int i;
  int j;
  int k;
  int num_entries;
  int opt;
  int tmp_int;
  FILE *cdb_file;
  char *target_cdb_path;
  char *alphabet = ALPHABET;
  struct main_index_record main_index[CDB_INDEX_COUNT];
  struct record *record_list;
  int tmp_slot_num;
  struct cdb_index tmp_cdb_index;
  struct cdb_hash_slot tmp_cdb_hash_slot;
  struct cdb_record_header tmp_cdb_record_header;
  long current_offset;
  int verbose;
  int total_overlength_key;
  int total_hash_index_mismatch;
  int total_hash_key_mismatch;
  int total_index_offset_oversize;
  int total_hash_offset_oversize;
  int total_key_length_oversize;
  int total_data_length_oversize;
  int total_corrupt;
  long total_payload_bytes;

  target_cdb_path = NULL;
  num_entries = 0;
  record_list = NULL;
  verbose = 0;
  total_overlength_key = 0;
  total_hash_index_mismatch = 0;
  total_hash_key_mismatch = 0;
  total_index_offset_oversize = 0;
  total_hash_offset_oversize = 0;
  total_key_length_oversize = 0;
  total_data_length_oversize = 0;
  total_corrupt = 0;
  total_payload_bytes = 0;

  while ((opt = getopt(argc, argv, "f:n:vd:k:h:i:m:x:o:")) != -1)
    switch (opt)
      {
      case 'f':
        target_cdb_path = optarg;
        break;
      case 'n':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          num_entries = tmp_int;

        break;
      case 'v':
        verbose++;
        break;
      case 'o':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_overlength_key = tmp_int;

        break;
      case 'x':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_hash_index_mismatch = tmp_int;

        break;
      case 'm':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_hash_key_mismatch = tmp_int;

        break;
      case 'i':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_index_offset_oversize = tmp_int;

        break;
      case 'h':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_hash_offset_oversize = tmp_int;

        break;
      case 'k':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_key_length_oversize = tmp_int;

        break;
      case 'd':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          total_data_length_oversize = tmp_int;

        break;
      default:
        usage();
        break;
      }

  total_corrupt = total_overlength_key + total_hash_index_mismatch + total_hash_key_mismatch + total_index_offset_oversize + total_hash_offset_oversize + total_key_length_oversize + total_data_length_oversize;
  if (total_corrupt > num_entries)
    {
    if (verbose)
      printf("Total number of corrupt entries is higher than the total number of entries, adjusting...\n");
    if (total_overlength_key > 0)
      {
      total_overlength_key = (num_entries * total_overlength_key) / total_corrupt;
      if (verbose)
        printf("Adjusting overlength keys to %d\n", total_overlength_key);
      }
    if (total_hash_index_mismatch > 0)
      {
      total_hash_index_mismatch = (num_entries * total_hash_index_mismatch) / total_corrupt;
      if (verbose)
        printf("Adjusting hash/index mismatches to %d\n", total_hash_index_mismatch);
      }
    if (total_hash_key_mismatch > 0)
      {
      total_hash_key_mismatch = (num_entries * total_hash_key_mismatch) / total_corrupt;
      if (verbose)
        printf("Adjusting hash/key mismatches to %d\n", total_hash_key_mismatch);
      }
    if (total_index_offset_oversize > 0)
      {
      total_index_offset_oversize = (num_entries * total_index_offset_oversize) / total_corrupt;
      if (verbose)
        printf("Adjusting oversize index offsets to %d\n", total_index_offset_oversize);
      }
    if (total_hash_offset_oversize > 0)
      {
      total_hash_offset_oversize = (num_entries * total_hash_offset_oversize) / total_corrupt;
      if (verbose)
        printf("Adjusting oversize hash offsets to %d\n", total_hash_offset_oversize);
      }
    if (total_key_length_oversize > 0)
      {
      total_key_length_oversize = (num_entries * total_key_length_oversize) / total_corrupt;
      if (verbose)
        printf("Adjusting oversize key lengths to %d\n", total_key_length_oversize);
      }
    if (total_data_length_oversize > 0)
      {
      total_data_length_oversize = (num_entries * total_data_length_oversize) / total_corrupt;
      if (verbose)
        printf("Adjusting oversize data lengths to %d\n", total_data_length_oversize);
      }
    }

  if (target_cdb_path != NULL)
    if ((cdb_file = fopen(target_cdb_path, "w")) != NULL)
      if ((num_entries == 0) ||
          ((record_list = malloc(sizeof(struct record) * num_entries)) != NULL))
        {
        srandom(time(NULL));

        for (i = 0; i < CDB_INDEX_COUNT; i++)
          {
          main_index[i].hash_values = NULL;
          main_index[i].num_slots = 0;
          main_index[i].slots = NULL;
          main_index[i].corrupt_index_offset_oversize = 0;
          }

        for (i = 0; i < num_entries; i++)
          {
          record_list[i].corrupt_overlength_key = 0;
          record_list[i].corrupt_hash_index_mismatch = 0;
          record_list[i].corrupt_hash_key_mismatch = 0;
          record_list[i].corrupt_index_offset_oversize = 0;
          record_list[i].corrupt_hash_offset_oversize = 0;
          record_list[i].corrupt_key_length_oversize = 0;
          record_list[i].corrupt_data_length_oversize = 0;
          }

        tmp_int = 0;
        for (i = 0; i < total_overlength_key; i++)
          record_list[tmp_int++].corrupt_overlength_key = 1;
        for (i = 0; i < total_hash_index_mismatch; i++)
          record_list[tmp_int++].corrupt_hash_index_mismatch = 1;
        for (i = 0; i < total_hash_key_mismatch; i++)
          record_list[tmp_int++].corrupt_hash_key_mismatch = 1;
        for (i = 0; i < total_index_offset_oversize; i++)
          record_list[tmp_int++].corrupt_index_offset_oversize = 1;
        for (i = 0; i < total_hash_offset_oversize; i++)
          record_list[tmp_int++].corrupt_hash_offset_oversize = 1;
        for (i = 0; i < total_key_length_oversize; i++)
          record_list[tmp_int++].corrupt_key_length_oversize = 1;
        for (i = 0; i < total_data_length_oversize; i++)
          record_list[tmp_int++].corrupt_data_length_oversize = 1;

        for (i = 0; i < num_entries; i++)
          {
          record_list[i].strlen_key = ((record_list[i].corrupt_overlength_key) ? MAX_BUF : 0) + (random() % MAX_BUF);
          for (j = 0; j < record_list[i].strlen_key; j++)
            record_list[i].key[j] = alphabet[random() % STRLEN(ALPHABET)];
          record_list[i].key[j] = '\0';

          record_list[i].hash = cdb_hash(record_list[i].key, record_list[i].strlen_key);

          if (record_list[i].corrupt_hash_key_mismatch)
            {
            for (j = 0; j < record_list[i].strlen_key; j++)
              record_list[i].key[j] = alphabet[random() % STRLEN(ALPHABET)];
            record_list[i].key[j] = '\0';
            }

          record_list[i].strlen_data = random() % MAX_BUF;
          for (j = 0; j < record_list[i].strlen_data; j++)
            record_list[i].data[j] = alphabet[random() % STRLEN(ALPHABET)];
          record_list[i].data[j] = '\0';

          record_list[i].index = record_list[i].hash % CDB_INDEX_COUNT;
          if (record_list[i].corrupt_hash_index_mismatch)
            {
            do
              {
              tmp_int = random() % CDB_INDEX_COUNT;
              }
            while (record_list[i].index == tmp_int);
            record_list[i].index = tmp_int;
            }

          record_list[i].slot_num = -1;

          if ((main_index[record_list[i].index].hash_values = realloc(main_index[record_list[i].index].hash_values, sizeof(uint32_t) * (main_index[record_list[i].index].num_slots + 1))) != NULL)
            {
            main_index[record_list[i].index].hash_values[main_index[record_list[i].index].num_slots] = record_list[i].hash;
            main_index[record_list[i].index].num_slots++;

            if (record_list[i].corrupt_index_offset_oversize)
              main_index[record_list[i].index].corrupt_index_offset_oversize = 1;

            if (verbose >= 2)
              printf("Generated record %d of %d: hash = %u, index = %d, hash values within index = %d\n", i, num_entries, record_list[i].hash, record_list[i].index, main_index[record_list[i].index].num_slots);
            }
          else
            {
            printf("ERROR: Unable to allocate %ld bytes\n", (long int)sizeof(uint32_t) * (main_index[record_list[i].index].num_slots + 1));
            exit(0);
            }

          total_payload_bytes += record_list[i].strlen_key + record_list[i].strlen_data;
          }

        for (i = 0; i < CDB_INDEX_COUNT; i++)
          if ((main_index[i].slots = (int *)malloc(sizeof(int) * main_index[i].num_slots)) != NULL)
            for (j = 0; j < main_index[i].num_slots; j++)
              main_index[i].slots[j] = 0;
          else
            {
            printf("ERROR: Unable to allocate %ld bytes\n", (long int)sizeof(int) * main_index[i].num_slots);
            exit(0);
            }

        current_offset = sizeof(struct cdb_index) * CDB_INDEX_COUNT;
        for (i = 0; i < CDB_INDEX_COUNT; i++)
          {
          tmp_cdb_index.offset = (main_index[i].num_slots == 0) ? htole(0) : htole(current_offset + (main_index[i].corrupt_index_offset_oversize ? ((sizeof(struct cdb_index) * CDB_INDEX_COUNT) + ((sizeof(struct cdb_hash_slot) + sizeof(struct cdb_record_header)) * num_entries) + total_payload_bytes) : 0));
          tmp_cdb_index.num_slots = htole(main_index[i].num_slots);

          if (fwrite(&tmp_cdb_index, sizeof(struct cdb_index), 1, cdb_file) == 1)
            {
            if (verbose)
              printf("Writing main index %d, offset = %u, num_slots = %d\n", i, htole(tmp_cdb_index.offset), main_index[i].num_slots);
            }
          else
            {
            printf("ERROR: Unable to write %ld bytes to %s\n", (long int)sizeof(struct cdb_index), target_cdb_path);
            exit(0);
            }

          current_offset += (sizeof(struct cdb_hash_slot) * main_index[i].num_slots);
          }

        for (i = 0; i < num_entries; i++)
          {
          tmp_slot_num = (record_list[i].hash / 256) % main_index[record_list[i].index].num_slots;
          if (main_index[record_list[i].index].slots[tmp_slot_num] == 0)
            record_list[i].slot_num = tmp_slot_num;
          else
            {
            tmp_slot_num++;
            if (tmp_slot_num == main_index[record_list[i].index].num_slots)
              tmp_slot_num = 0;
            }
          }

        for (i = 0; i < CDB_INDEX_COUNT; i++)
          for (j = 0; j < main_index[i].num_slots; j++)
            for (k = 0; k < num_entries; k++)
              if ((record_list[k].index == i) &&
                  (record_list[k].slot_num == j))
                {
                tmp_cdb_hash_slot.hash_value = htole(record_list[k].hash);
                tmp_cdb_hash_slot.offset = htole(current_offset + (record_list[k].corrupt_hash_offset_oversize ? ((sizeof(struct cdb_index) * CDB_INDEX_COUNT) + ((sizeof(struct cdb_hash_slot) + sizeof(struct cdb_record_header)) * num_entries) + total_payload_bytes) : 0));

                if (fwrite(&tmp_cdb_hash_slot, sizeof(struct cdb_hash_slot), 1, cdb_file) == 1)
                  {
                  if (verbose)
                    printf("Writing secondary index at offset %ld: index = %d, slot = %d, hash = %u, offset = %d\n", ftell(cdb_file) - (long int)sizeof(struct cdb_hash_slot), i, j, htole(tmp_cdb_hash_slot.hash_value), htole(tmp_cdb_hash_slot.offset));
                  }
                else
                  {
                  printf("ERROR: Unable to write %ld bytes to %s\n", (long int)sizeof(struct cdb_hash_slot), target_cdb_path);
                  exit(0);
                  }

                current_offset += sizeof(struct cdb_record_header) + record_list[k].strlen_key + record_list[k].strlen_data;
                }

        for (i = 0; i < CDB_INDEX_COUNT; i++)
          for (j = 0; j < main_index[i].num_slots; j++)
            for (k = 0; k < num_entries; k++)
              if ((record_list[k].index == i) &&
                  (record_list[k].slot_num == j))
                {
                tmp_cdb_record_header.key_length = htole(record_list[k].strlen_key + (record_list[k].corrupt_key_length_oversize ? ((sizeof(struct cdb_index) * CDB_INDEX_COUNT) + ((sizeof(struct cdb_hash_slot) + sizeof(struct cdb_record_header)) * num_entries) + total_payload_bytes) : 0));
                tmp_cdb_record_header.data_length = htole(record_list[k].strlen_data + (record_list[k].corrupt_data_length_oversize ? ((sizeof(struct cdb_index) * CDB_INDEX_COUNT) + ((sizeof(struct cdb_hash_slot) + sizeof(struct cdb_record_header)) * num_entries) + total_payload_bytes) : 0));

                if (fwrite(&tmp_cdb_record_header, sizeof(struct cdb_record_header), 1, cdb_file) == 1)
                  {
                  if (verbose)
                    printf("Writing data record at offset %ld: index = %d, hash = %u, key_length = %d, data_length = %d\n", ftell(cdb_file) - (long int)sizeof(struct cdb_record_header), record_list[k].index, record_list[k].hash, htole(tmp_cdb_record_header.key_length), htole(tmp_cdb_record_header.data_length));
                  }
                else
                  {
                  printf("ERROR: Unable to write %ld bytes to %s\n", (long int)sizeof(struct cdb_record_header), target_cdb_path);
                  exit(0);
                  }

                if (fwrite(record_list[k].key, sizeof(char), record_list[k].strlen_key, cdb_file) != record_list[k].strlen_key)
                  {
                  printf("ERROR: Unable to write %ld bytes to %s\n", (long int)sizeof(char) * record_list[k].strlen_key, target_cdb_path);
                  exit(0);
                  }

                if (fwrite(record_list[k].data, sizeof(char), record_list[k].strlen_data, cdb_file) != record_list[k].strlen_data)
                  {
                  printf("ERROR: Unable to write %ld bytes to %s\n", (long int)sizeof(char) * record_list[k].strlen_data, target_cdb_path);
                  exit(0);
                  }
                }
        }
      else
        printf("ERROR: Unable to allocate %ld of memory\n", (long int)sizeof(struct record) * num_entries);
    else
      printf("ERROR: Unable to open %s for writing: %s\n", target_cdb_path, strerror(errno));
  else
    usage();

  return(0);
  }
