/*
  generator -- a program for creating spamdyke test scripts
  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * This program recursively explores every possible code path in spamdyke's
 * recipient validation algorithm (see the flowchart in the documentation
 * folder for details).  Because there are hundreds of thousands of possible
 * paths, there's simply no way to create the scripts by hand.  In addition,
 * including all those scripts in the spamdyke tarball would increase its
 * size significantly.
 *
 * Each test is created in its own folder and follows the structure expected
 * by the "run" script.  The script names are generated using the flowchart
 * path they represent.  For example, the name "YYYNNYN" means the script
 * follows the path Start->Yes->Yes->Yes->No->No->Yes->No in the flowchart.
 * The comments at the top of each script provide more detail about the
 * conditions it creates before qmail is run.
 *
 * Two versions of each script are generated -- one that runs qmail alone and
 * one that runs qmail with spamdyke.  The idea is to test the flowchart's
 * correctness against qmail's actual behavior before testing spamdyke.
 * spamdyke shouldn't reject messages qmail would normally accept and
 * vice-versa.
 *
 * To make the code easier to match to the flowchart, every function in this
 * program begins with a number and two underscores (e.g. f15__).  The numbers
 * are shown on the flowchart in orange circles next to the decision nodes
 * they implement.
 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <math.h>
#include <errno.h>

#define MAX_BUF                         65535
#define STRLEN(X)                       ((int)(sizeof(X) - 1))
#define _STRINGIFY(X)                   #X
#define STRINGIFY(X)                    _STRINGIFY(X)
#define MINVAL(a,b)                     ({ typeof (a) _a = (a); typeof (b) _b = (b); _a < _b ? _a : _b; })
#define MAXVAL(a,b)                     ({ typeof (a) _a = (a); typeof (b) _b = (b); _a > _b ? _a : _b; })
#define PREFIX_SIZE                     1000
/*
 * The letters e and n are deliberately excluded because sometimes it happens
 * they are manipulated into a string that reads "-e" or "-n", which has
 * special meaning when passed to the "echo" command and makes it malfunction.
 * <sigh>
 */
#define RANDOM_ALPHABET                 "abcdfghijklmopqrstuvwxyz0123456789"

#define REJECTION_UNQUALIFIED           0
#define REJECTION_RELAYING              1
#define REJECTION_UNAVAILABLE           2
#define REJECTION_INVALID               3
#define REJECTION_CANNOT_START          4

#define ACCEPT_DELIVER                  0
#define ACCEPT_BOUNCE                   1
#define ACCEPT_DISCARD                  2
#define ACCEPT_REMOTE                   3

#define DECISION(FUNC_NAME,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,N_COMMENT,N_PRE,N_ACTION,N_POST) \
  DECISION_LIMITED(FUNC_NAME,-1,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,-1,N_COMMENT,N_PRE,N_ACTION,N_POST)
#define DECISION_LIMITED(FUNC_NAME,MAX_REENTER_Y,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,MAX_REENTER_N,N_COMMENT,N_PRE,N_ACTION,N_POST) \
  void FUNC_NAME() \
    { \
    static int reentered_y = 0; \
    static int reentered_n = 0; \
    static char tmp_post[MAX_BUF + 1]; \
    int tmp_strlen_pre; \
    int tmp_strlen_post; \
    int tmp_strlen; \
    if (verbose >= 2) \
      { \
      if (count_only) \
        printf("Counting total scripts %d: %s\n", total_number, STRINGIFY(FUNC_NAME)); \
      else \
        printf("Script %d: %s\n", next_number, STRINGIFY(FUNC_NAME)); \
      fflush(NULL); \
      } \
    if ((MAX_REENTER_N == -1) || \
        (MAX_REENTER_N > reentered_n)) \
      { \
      tmp_strlen_pre = 0; \
      tmp_strlen_post = 0; \
      reentered_n++; \
      path[strlen_path] = 'N'; \
      strlen_path++; \
      path[strlen_path] = '\0'; \
      snprintf(comments + strlen_comments, MAX_BUF - strlen_comments, "#   %s (%s)\n%n", N_COMMENT, STRINGIFY(FUNC_NAME), &tmp_strlen); \
      strlen_comments += tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      if (STRLEN(N_PRE) > 0) \
        { \
        snprintf(pre + strlen_pre, MAX_BUF - strlen_pre, "\n################################################################################\n# %s (%s)\nLEVEL_%s=%d\n%s%n", N_COMMENT, STRINGIFY(FUNC_NAME), STRINGIFY(FUNC_NAME), reentered_y + reentered_n, randomize(N_PRE, STRLEN(N_PRE)), &tmp_strlen_pre); \
        strlen_pre += tmp_strlen_pre; \
        pre[strlen_pre] = '\0'; \
        } \
      if (STRLEN(N_POST) > 0) \
        { \
        snprintf(tmp_post, MAX_BUF, "################################################################################\n# %s (%s)\n%sLEVEL_%s=%d\n\n%n", N_COMMENT, STRINGIFY(FUNC_NAME), randomize(N_POST, STRLEN(N_POST)), STRINGIFY(FUNC_NAME), reentered_y + (reentered_n - 1), &tmp_strlen_post); \
        memmove(post + tmp_strlen_post, post, strlen_post); \
        memcpy(post, tmp_post, tmp_strlen_post); \
        strlen_post += tmp_strlen_post; \
        post[strlen_post] = '\0'; \
        } \
      N_ACTION \
      strlen_comments -= tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      strlen_pre -= tmp_strlen_pre; \
      pre[strlen_pre] = '\0'; \
      if (STRLEN(N_POST) > 0) \
        { \
        strlen_post -= tmp_strlen_post; \
        memmove(post, post + tmp_strlen_post, strlen_post); \
        } \
      post[strlen_post] = '\0'; \
      strlen_path--; \
      path[strlen_path] = '\0'; \
      reentered_n--; \
      } \
    if ((MAX_REENTER_Y == -1) || \
        (MAX_REENTER_Y > reentered_y)) \
      { \
      tmp_strlen_pre = 0; \
      tmp_strlen_post = 0; \
      reentered_y++; \
      path[strlen_path] = 'Y'; \
      strlen_path++; \
      path[strlen_path] = '\0'; \
      snprintf(comments + strlen_comments, MAX_BUF - strlen_comments, "#   %s (%s)\n%n", Y_COMMENT, STRINGIFY(FUNC_NAME), &tmp_strlen); \
      strlen_comments += tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      if (STRLEN(Y_PRE) > 0) \
        { \
        snprintf(pre + strlen_pre, MAX_BUF - strlen_pre, "\n################################################################################\n# %s (%s)\nLEVEL_%s=%d\n%s%n", Y_COMMENT, STRINGIFY(FUNC_NAME), STRINGIFY(FUNC_NAME), reentered_y + reentered_n, randomize(Y_PRE, STRLEN(Y_PRE)), &tmp_strlen_pre); \
        strlen_pre += tmp_strlen_pre; \
        pre[strlen_pre] = '\0'; \
        } \
      if (STRLEN(Y_POST) > 0) \
        { \
        snprintf(tmp_post, MAX_BUF, "################################################################################\n# %s (%s)\n%sLEVEL_%s=%d\n\n%n", Y_COMMENT, STRINGIFY(FUNC_NAME), randomize(Y_POST, STRLEN(Y_POST)), STRINGIFY(FUNC_NAME), (reentered_y - 1) + reentered_n, &tmp_strlen_post); \
        memmove(post + tmp_strlen_post, post, strlen_post); \
        memcpy(post, tmp_post, tmp_strlen_post); \
        strlen_post += tmp_strlen_post; \
        post[strlen_post] = '\0'; \
        } \
      Y_ACTION \
      strlen_comments -= tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      strlen_pre -= tmp_strlen_pre; \
      pre[strlen_pre] = '\0'; \
      if (STRLEN(Y_POST) > 0) \
        { \
        strlen_post -= tmp_strlen_post; \
        memmove(post, post + tmp_strlen_post, strlen_post); \
        } \
      post[strlen_post] = '\0'; \
      strlen_path--; \
      path[strlen_path] = '\0'; \
      reentered_y--; \
      } \
    return; \
    }

int count_only;
int total_number;
int next_number;
long start_time;
int strlen_comments;
char comments[MAX_BUF + 1];
int strlen_path;
char path[MAX_BUF + 1];
int spamdyke;
char test_name[MAX_BUF + 1];
char test_dir[MAX_BUF + 1];
int strlen_pre;
char pre[MAX_BUF + 1];
int strlen_post;
char post[MAX_BUF + 1];
int max_strlen;
int restart_needed;
int relayclient_set;
int num_ext_hyphens;
int num_username_hyphens;
int username_truncated;
int using_defaultdelivery;
int dash_emptyable;
int dash_empty;
int envnoathost_exists;
int forwards_only;
int lines_in_file;
int comments_in_file;
int begin_num;
int end_num;
char target_prefix[MAX_BUF + 1];
int strlen_target_prefix;
int verbose;

char *alphabet = RANDOM_ALPHABET;
int last_even_letter = -1;
int even_direction = 1;
int last_odd_letter = -1;
int odd_direction = -1;

/*
 * This isn't really very random at all, but a lot of the generated tests will
 * fail if the same random domain names are generated in multiple places within
 * the same script.  Some of the string lengths are so constrained that there
 * are only two random characters in the domain name, so the odds of getting
 * the same two letters in a few scripts out of 100K is high enough that it
 * happens every time.
 *
 * So to try to prevent that, this function does more than just pick purely
 * randomly.  There's randomness involved in picking the start point, but after
 * that it's just sequential iteration.
 */
char *randomize(char *input_text, int strlen_input_text)
  {
  static char return_value[MAX_BUF + 1];
  int i;
  int count;

  if (last_even_letter == -1)
    last_even_letter = random() % STRLEN(RANDOM_ALPHABET);
  if (last_odd_letter == -1)
    {
    last_odd_letter = last_even_letter + (STRLEN(RANDOM_ALPHABET) / 2);
    if (last_odd_letter >= STRLEN(RANDOM_ALPHABET))
      last_odd_letter -= STRLEN(RANDOM_ALPHABET);
    }

  count = 0;
  for (i = 0; i < strlen_input_text; i++)
    if (input_text[i] == '\0')
      {
      if ((count % 2) == 0)
        return_value[i] = alphabet[last_odd_letter];
      else
        {
        return_value[i] = alphabet[last_even_letter];
        last_even_letter += even_direction;
        if ((last_even_letter == STRLEN(RANDOM_ALPHABET)) ||
            (last_even_letter < 0))
          {
          even_direction *= -1;
          last_even_letter += even_direction;

          last_odd_letter += odd_direction;
          if ((last_odd_letter == STRLEN(RANDOM_ALPHABET)) ||
              (last_odd_letter < 0))
            {
            odd_direction *= -1;
            last_odd_letter += odd_direction;
            }
          }
        }

      count++;
      }
    else
      return_value[i] = input_text[i];
  return_value[i] = '\0';

  return(return_value);
  }

FILE *open_file(char *target_path, char *accept_reject, char *suffix)
  {
  static char prefix_format[512] = { '\0' };
  static char filename_format[512] = { '\0' };
  FILE *return_value;
  int tmp_strlen;
  long secs_remaining;
  long secs_elapsed;
  char prefix_dir[MAX_BUF + 1];
  struct stat tmp_stat;

  return_value = NULL;

  if (prefix_format[0] == '\0')
    snprintf(prefix_format, 512, "%%s/%%.%dd", (int)ceil(log10(total_number)));
  if (filename_format[0] == '\0')
    snprintf(filename_format, 512, "%%s/%%.%dd/%%.%dd-%%s_%%s%%s%%n", (int)ceil(log10(total_number)), (int)ceil(log10(total_number)));

  snprintf(prefix_dir, MAX_BUF, prefix_format, test_dir, next_number - (next_number % PREFIX_SIZE));
  snprintf(test_name, MAX_BUF, filename_format, test_dir, next_number - (next_number % PREFIX_SIZE), next_number, target_path, accept_reject, suffix, &tmp_strlen);

  printf("Creating test %d of %d: %s", next_number, total_number, test_name);
  if ((next_number < 100) ||
      ((next_number % 20) != 0))
    printf("\n");
  else
    {
    secs_elapsed = time(NULL) - start_time;
    if (secs_elapsed >= 180)
      {
      secs_remaining = (secs_elapsed * (total_number - next_number)) / next_number;
      printf(": \t%ld:%.2ld:%.2ld elapsed, estimated %ld:%.2ld:%.2ld remaining\n", secs_elapsed / 3600, (secs_elapsed % 3600) / 60, (secs_elapsed % 3600) % 60, secs_remaining / 3600, (secs_remaining % 3600) / 60, (secs_remaining % 3600) % 60);
      }
    else
      printf("\n");
    }

  if (stat(prefix_dir, &tmp_stat) == 0)
    {
    if (verbose)
      printf("Found directory %s\n", prefix_dir);

    if ((tmp_stat.st_mode & S_IFDIR) != S_IFDIR)
      {
      fprintf(stderr, "Non-directory already exists where directory should be: %s\n", prefix_dir);
      exit(1);
      }
    }
  else if (errno == ENOENT)
    {
    if (verbose)
      printf("Creating directory %s\n", prefix_dir);

    if (mkdir(prefix_dir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1)
      {
      fprintf(stderr, "Cannot mkdir %s: %s\n", prefix_dir, strerror(errno));
      exit(1);
      }
    }
  else
    {
    fprintf(stderr, "Cannot stat %s: %s\n", prefix_dir, strerror(errno));
    exit(1);
    }

  if (verbose)
    printf("Creating directory %s\n", test_name);

  if (mkdir(test_name, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1)
    {
    fprintf(stderr, "Cannot mkdir %s: %s\n", test_name, strerror(errno));
    exit(1);
    }

  snprintf(test_name + tmp_strlen, MAX_BUF - tmp_strlen, "/run.sh");

  if (verbose)
    printf("Creating script %s\n", test_name);

  if ((return_value = fopen(test_name, "w")) == NULL)
    {
    fprintf(stderr, "Cannot fopen %s: %s\n", test_name, strerror(errno));
    exit(1);
    }

  return(return_value);
  }

#define REJECT_TOP \
  "# The message should be rejected\n" \
  "\n" \
  "export TCPREMOTEIP=11.22.33.44\n" \
  "unset GREP_TARGET\n" \
  "unset REJECT_GREP_TARGET\n" \
  "unset TEMP_FILENAME\n" \
  "unset HOMEDIR\n" \
  "unset DASH\n" \
  "unset EXT\n" \
  "unset REAL_RELAYCLIENT\n" \
  "unset REAL_USERNAME\n" \
  "unset REAL_DOMAIN\n" \
  "unset REAL_ADDRESS\n" \
  "unset TO_USERNAME\n" \
  "unset TO_DOMAIN\n" \
  "unset TO_ADDRESS\n" \
  "unset TEMP_USERNAME\n" \
  "unset WILDCARD_DOT\n" \
  "unset WILDCARD_DASH\n" \
  "unset WILDCARD_DASH2\n" \
  "unset STRLEN_REAL_USERNAME\n" \
  "unset PATTERN\n" \
  "unset WILDCARD_ADDITION\n" \
  "unset IN_MORERCPTHOSTS\n" \
  "unset IN_RCPTHOSTS\n" \
  "unset PERCENTHACK_USERNAME_ADDITION\n" \
  "unset VIRTUALDOMAINS_USERNAME_ADDITION\n" \
  "FROM_ADDRESS=test-${TEST_NUM}.\0\0\0\0\0.\0\0\0\0\0@example.com\n" \
  "REAL_USERNAME=s-\0\0\0\n" \
  "REAL_DOMAIN=\0\0.test\n" \
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n" \
  "TO_USERNAME=${REAL_USERNAME}\n" \
  "TO_DOMAIN=${REAL_DOMAIN}\n" \
  "TO_ADDRESS=${REAL_ADDRESS}\n" \
  "TEMP_USERNAME=${REAL_USERNAME}\n" \
  "VALID_DOMAIN=valid-domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n" \
  "VALID_USERNAME=valid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n" \
  "VALID_ADDRESS=${VALID_USERNAME}@${VALID_DOMAIN}\n" \
  "INVALID_USERNAME=invalid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n" \
  "INVALID_ADDRESS=${INVALID_USERNAME}@${VALID_DOMAIN}\n" \
  "VALID_HOMEDIR=${TMPDIR}/${TEST_NUM}-valid_home\n" \
  "VALID_GREP_TARGET=${VALID_HOMEDIR}/mbox\n" \
  "mkdir -p ${VALID_HOMEDIR}\n" \
  "echo ${VALID_HOMEDIR}/mbox > ${VALID_HOMEDIR}/.qmail\n" \
  "touch ${VALID_HOMEDIR}/mbox\n" \
  "sync\n" \
  "chmod 700 ${VALID_HOMEDIR}\n" \
  "chown -R $6 ${VALID_HOMEDIR}\n" \
  "chgrp -R $7 ${VALID_HOMEDIR}\n" \
  "chmod 600 ${VALID_HOMEDIR}/mbox\n" \
  "chown $6 ${VALID_HOMEDIR}/mbox\n" \
  "chgrp $7 ${VALID_HOMEDIR}/mbox\n" \
  "\n" \
  "%s" \
  "\n" \
  "################################################################################\n" \
  "# final steps\n" \
  "if [ -z \"${GREP_TARGET}\" ]\n" \
  "then\n" \
  "  GREP_TARGET=${REJECT_GREP_TARGET}\n" \
  "fi\n"

void reject_message(int rejection_type)
  {
  FILE *script_file;

  if (!count_only)
    {
    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        (strlen_target_prefix == 0))
      {
      if ((next_number % 1000) == 0)
        {
        script_file = open_file("X", "clean_tmp", "");

        fprintf(script_file,
          "# This script was automatically generated by the \"generator\" program.\n"
          "# Any bugs should be fixed there, not here.\n"
          "#\n"
          "# This script deletes all the files from the temporary folder because the\n"
          "# folder will fill up eventually and no more files can be created, which\n"
          "# will cause later scripts to fail.\n"
          "\n"
          "TEST_FOLDER=`echo ${TEST_NUM} | sed -e \"s/_[0-9]*$//\"`\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -print0 2> /dev/null | xargs -0 chmod -R 777\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -type d -print0 2> /dev/null | xargs -0 rm -rf\n"
          "\n"
          "outcome=\"success\"\n"
          );

        fclose(script_file);
        }
      }

    if ((next_number % 1000) == 0)
      next_number++;

    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        ((strlen_target_prefix == 0) ||
         !strncmp(target_prefix, path, strlen_target_prefix)))
      {
      script_file = open_file(path, "reject", spamdyke ? "_spamdyke" : "_qmail_only");

      fprintf(script_file,
        "# This script was automatically generated by the \"generator\" program.\n"
        "# Any bugs should be fixed there, not here.\n"
        "#\n"
        "# This script tests recipient validation using the flowchart in the\n"
        "# documentation folder.  It follows this path through the flowchart:\n"
        "%s",
        comments);

      fprintf(script_file,
        randomize(REJECT_TOP, STRLEN(REJECT_TOP)),
        pre);

      fprintf(script_file,
        "\n"
        "################################################################################\n"
        "# empty queue\n"
        "echo Emptying queue and restarting qmail...\n"
        "svc -d /service/qmail-send\n"
        "svc -d /service/qmail-smtpd\n"
        "for i in 2 1 0\n"
        "do\n"
        "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
        "  if [ -z \"${output}\" ]\n"
        "  then\n"
        "    break\n"
        "  else\n"
        "    ${SLEEP_PATH} 1\n"
        "  fi\n"
        "done\n"
        "if [ ! -z \"${output}\" ]\n"
        "then\n"
        "  svc -k /service/qmail-send\n"
        "  svc -k /service/qmail-smtpd\n"
        "fi\n"
        "rm -f `find /var/qmail/queue/mess /var/qmail/queue/info /var/qmail/queue/local /var/qmail/queue/remote /var/qmail/queue/bounce /var/qmail/queue/intd /var/qmail/queue/todo -type f -print`\n"
        "svc -u /service/qmail-send\n"
        "svc -u /service/qmail-smtpd\n");

      fprintf(script_file,
        "\n"
        "\n"
        "################################################################################\n"
        "# create input file\n"
        "echo \"HELO me\" > ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"MAIL FROM:<${FROM_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"RCPT TO:<${TO_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n");

      if (!spamdyke)
        fprintf(script_file,
          "echo \"DATA\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
          "echo \"To: <${REAL_ADDRESS}>\\r\\nFrom:<${FROM_ADDRESS}>\\r\\nSubject: ${TEST_NUM}\\r\\n\\r\\nFOO!\\r\\n.\\r\\n\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
          );

      fprintf(script_file,
        "echo \"QUIT\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "\n"
        );

      if (spamdyke)
        {
        fprintf(script_file,
          "################################################################################\n"
          "# run test\n"
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "sync\n"
          "\n"
          );
        switch (rejection_type)
          {
          case REJECTION_UNQUALIFIED:
            fprintf(script_file,
              "output=`grep \"554 Improper recipient address. Try supplying a domain name.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "if [ ! -z \"${output}\" ]\n"
              "then\n"
              "  output=`grep \"DENIED_UNQUALIFIED_RECIPIENT from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_RELAYING:
            fprintf(script_file,
              "output=`grep \"554 Refused. Sending to remote addresses (relaying) is not allowed.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "if [ ! -z \"${output}\" ]\n"
              "then\n"
              "  output=`grep \"DENIED_RELAYING from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_UNAVAILABLE:
            fprintf(script_file,
              "output=`grep \"554 Refused. The recipient is not accepting mail right now.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "if [ ! -z \"${output}\" ]\n"
              "then\n"
              "  output=`grep \"DENIED_UNAVAILABLE_RECIPIENT from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_INVALID:
            fprintf(script_file,
              "output=`grep \"554 Refused. The recipient address does not exist.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "if [ ! -z \"${output}\" ]\n"
              "then\n"
              "  output=`grep \"DENIED_INVALID_RECIPIENT from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_CANNOT_START:
            printf("ERROR: reject_message() was passed REJECTION_CANNOT_START when spamdyke was true.  This should not be possible.\n");
            exit(0);
            break;
          }
        fprintf(script_file,
          "  if [ ! -z \"${output}\" ]\n"
          "  then\n"
          "    outcome=\"success\"\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n"
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );
        }
      else
        {
        fprintf(script_file,
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "sync\n"
          "\n"
          );
        if (rejection_type == REJECTION_CANNOT_START)
          fprintf(script_file,
            "output=`grep \"421 unable to read controls (#4.3.0)\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
            "if [ ! -z \"${output}\" ]\n"
            "then\n"
            "  outcome=\"success\"\n"
            "else\n"
            "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
            "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
            "\n"
            "  outcome=\"failure\"\n"
            "fi\n"
            );
        else if (rejection_type == REJECTION_RELAYING)
          fprintf(script_file,
            "output=`grep \"553 sorry, that domain isn't in my list of allowed rcpthosts (#5.7.1)\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
            "if [ ! -z \"${output}\" ]\n"
            "then\n"
            "  outcome=\"success\"\n"
            "else\n"
            "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
            "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
            "\n"
            "  outcome=\"failure\"\n"
            "fi\n"
            );
        else
          fprintf(script_file,
            "unset outcome\n"
            "unset MAILDIR\n"
            "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
            "if [ ! -z \"${output}\" ]\n"
            "then\n"
            "  echo \"Waiting to see if qmail queues or bounces the message...\"\n"
            "  svc -a /service/qmail-send\n"
            "\n"
            "  for i in 10 9 8 7 6 5 4 3 2 1 0\n"
            "  do\n"
            "    sync\n"
            "    if [ -f ~postmaster/.qmail ]\n"
            "    then\n"
            "      MAILDIR=`cat ~postmaster/.qmail`\n"
            "      if [ \"${MAILDIR:0:1}\" != \"/\" ]\n"
            "      then\n"
            "        MAILDIR=~postmaster/${MAILDIR}\n"
            "      fi\n"
            "      output=`grep -r ${FROM_ADDRESS} ${MAILDIR}`\n"
            "      if [ ! -z \"${output}\" ]\n"
            "      then\n"
            "        echo \"Found bounced message in ${MAILDIR}\"\n"
            "        outcome=\"success\"\n"
            "        break\n"
            "      fi\n"
            "    fi\n"
            "    if [ -f /var/log/qmail/current ]\n"
            "    then\n"
            "      output=`grep -E \"starting delivery .: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
            "      if [ ! -z \"${output}\" ]\n"
            "      then\n"
            "        echo \"Found log of bounced message in /var/log/qmail/current\"\n"
            "        outcome=\"success\"\n"
            "        break\n"
            "      else\n"
            "        output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
            "        if [ ! -z \"${output}\" ]\n"
            "        then\n"
            "          echo \"Found queued message in /var/qmail/queue/local\"\n"
            "          outcome=\"success\"\n"
            "          break\n"
            "        else\n"
            "          output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
            "          if [ ! -z \"${output}\" ]\n"
            "          then\n"
            "            echo \"Found queued message in /var/qmail/queue/remote\"\n"
            "            outcome=\"success\"\n"
            "            break\n"
            "          fi\n"
            "        fi\n"
            "      fi\n"
            "    fi\n"
            "    ${SLEEP_PATH} 1\n"
            "  done\n"
            "\n"
            "  if [ -z \"${outcome}\" ]\n"
            "  then\n"
            "    if [ ! -z \"${MAILDIR}\" ]\n"
            "    then\n"
            "      echo \"COULD NOT FIND BOUNCED MESSAGE FROM ${FROM_ADDRESS} IN ${MAILDIR}; SHOULD HAVE BEEN BOUNCED\"\n"
            "    fi\n"
            "\n"
            "    echo \"COULD NOT FIND QUEUED MESSAGE TO ${REAL_ADDRESS} IN /var/qmail/queue/local OR TO ${FROM_ADDRESS} IN LOGS OR /var/qmail/queue/remote; SHOULD HAVE BEEN QUEUED OR BOUNCED\"\n"
            "    outcome=\"failure\"\n"
            "  fi\n"
            "else\n"
            "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
            "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
            "\n"
            "  outcome=\"failure\"\n"
            "fi\n"
            );
        }

      fprintf(script_file, "\n%s", post);

      if (restart_needed > 0)
        fprintf(script_file,
          "\n"
          "echo Restarting qmail...\n"
          "svc -d /service/qmail-send\n"
          "svc -d /service/qmail-smtpd\n"
          "for i in 2 1 0\n"
          "do\n"
          "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
          "  if [ -z \"${output}\" ]\n"
          "  then\n"
          "    break\n"
          "  else\n"
          "    ${SLEEP_PATH} 1\n"
          "  fi\n"
          "done\n"
          "if [ ! -z \"${output}\" ]\n"
          "then\n"
          "  svc -k /service/qmail-send\n"
          "  svc -k /service/qmail-smtpd\n"
          "fi\n"
          "svc -u /service/qmail-send\n"
          "svc -u /service/qmail-smtpd\n");

      fprintf(script_file, "\n");

      fclose(script_file);
      }

    next_number++;
    }
  else
    {
    if ((total_number % 1000) == 0)
      total_number++;

    total_number++;
    max_strlen = MAXVAL(max_strlen, MAXVAL(strlen_pre, strlen_post));
    }

  return;
  }

#define ACCEPT_TOP \
  "# The message should be accepted\n" \
  "\n" \
  "export TCPREMOTEIP=11.22.33.44\n" \
  "unset GREP_TARGET\n" \
  "unset REJECT_GREP_TARGET\n" \
  "unset TEMP_FILENAME\n" \
  "unset HOMEDIR\n" \
  "unset DASH\n" \
  "unset EXT\n" \
  "unset REAL_RELAYCLIENT\n" \
  "unset REAL_USERNAME\n" \
  "unset REAL_DOMAIN\n" \
  "unset REAL_ADDRESS\n" \
  "unset TO_USERNAME\n" \
  "unset TO_DOMAIN\n" \
  "unset TO_ADDRESS\n" \
  "unset TEMP_USERNAME\n" \
  "unset WILDCARD_DOT\n" \
  "unset WILDCARD_DASH\n" \
  "unset WILDCARD_DASH2\n" \
  "unset STRLEN_REAL_USERNAME\n" \
  "unset PATTERN\n" \
  "unset WILDCARD_ADDITION\n" \
  "unset IN_MORERCPTHOSTS\n" \
  "unset IN_RCPTHOSTS\n" \
  "unset PERCENTHACK_USERNAME_ADDITION\n" \
  "unset VIRTUALDOMAINS_USERNAME_ADDITION\n" \
  "FROM_ADDRESS=test-${TEST_NUM}.\0\0\0\0\0.\0\0\0\0\0@example.com\n" \
  "REAL_USERNAME=s-\0\0\0\n" \
  "REAL_DOMAIN=\0\0.test\n" \
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n" \
  "TO_USERNAME=${REAL_USERNAME}\n" \
  "TO_DOMAIN=${REAL_DOMAIN}\n" \
  "TO_ADDRESS=${REAL_ADDRESS}\n" \
  "TEMP_USERNAME=${REAL_USERNAME}\n" \
  "VALID_DOMAIN=domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n" \
  "VALID_USERNAME=user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n" \
  "VALID_ADDRESS=${VALID_USERNAME}@${VALID_DOMAIN}\n" \
  "VALID_HOMEDIR=${TMPDIR}/${TEST_NUM}-valid_home\n" \
  "VALID_GREP_TARGET=${VALID_HOMEDIR}/mbox\n" \
  "mkdir -p ${VALID_HOMEDIR}\n" \
  "chmod 700 ${VALID_HOMEDIR}\n" \
  "chown -R $6 ${VALID_HOMEDIR}\n" \
  "chgrp -R $7 ${VALID_HOMEDIR}\n" \
  "echo ${VALID_HOMEDIR}/mbox > ${VALID_HOMEDIR}/.qmail\n" \
  "touch ${VALID_HOMEDIR}/mbox\n" \
  "chmod 600 ${VALID_HOMEDIR}/mbox\n" \
  "chown $6 ${VALID_HOMEDIR}/mbox\n" \
  "chgrp $7 ${VALID_HOMEDIR}/mbox\n" \
  "\n" \
  "%s" \
  "\n" \
  "################################################################################\n" \
  "# final steps\n" \
  "if [ -z \"${GREP_TARGET}\" ]\n" \
  "then\n" \
  "  GREP_TARGET=${REJECT_GREP_TARGET}\n" \
  "fi\n"

void accept_message(int accept_type)
  {
  FILE *script_file;

  if (!count_only)
    {
    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        (strlen_target_prefix == 0))
      {
      if ((next_number % 1000) == 0)
        {
        script_file = open_file("X", "clean_tmp", "");

        fprintf(script_file,
          "# This script was automatically generated by the \"generator\" program.\n"
          "# Any bugs should be fixed there, not here.\n"
          "#\n"
          "# This script deletes all the files from the temporary folder because the\n"
          "# folder will fill up eventually and no more files can be created, which\n"
          "# will cause later scripts to fail.\n"
          "\n"
          "TEST_FOLDER=`echo ${TEST_NUM} | sed -e \"s/_[0-9]*$//\"`\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -print0 2> /dev/null | xargs -0 chmod 777\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -type d -print0 2> /dev/null | xargs -0 rm -rf\n"
          "\n"
          "outcome=\"success\"\n"
          );

        fclose(script_file);
        }
      }

    if ((next_number % 1000) == 0)
      next_number++;

    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        ((strlen_target_prefix == 0) ||
         !strncmp(target_prefix, path, strlen_target_prefix)))
      {
      script_file = open_file(path, "accept", spamdyke ? "_spamdyke" : "_qmail_only");

      fprintf(script_file,
        "# This script was automatically generated by the \"generator\" program.\n"
        "# Any bugs should be fixed there, not here.\n"
        "#\n"
        "# This script tests recipient validation using the flowchart in the\n"
        "# documentation folder.  It follows this path through the flowchart:\n"
        "%s",
        comments);

      fprintf(script_file,
        randomize(ACCEPT_TOP, STRLEN(ACCEPT_TOP)),
        pre);

      fprintf(script_file,
        "\n"
        "################################################################################\n"
        "# empty queue\n"
        "echo Emptying queue and restarting qmail...\n"
        "svc -d /service/qmail-send\n"
        "svc -d /service/qmail-smtpd\n"
        "for i in 2 1 0\n"
        "do\n"
        "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
        "  if [ -z \"${output}\" ]\n"
        "  then\n"
        "    break\n"
        "  else\n"
        "    ${SLEEP_PATH} 1\n"
        "  fi\n"
        "done\n"
        "if [ ! -z \"${output}\" ]\n"
        "then\n"
        "  svc -k /service/qmail-send\n"
        "  svc -k /service/qmail-smtpd\n"
        "fi\n"
        "rm -f `find /var/qmail/queue/mess /var/qmail/queue/info /var/qmail/queue/local /var/qmail/queue/remote /var/qmail/queue/bounce /var/qmail/queue/intd /var/qmail/queue/todo -type f -print`\n"
        "svc -u /service/qmail-send\n"
        "svc -u /service/qmail-smtpd\n");

      fprintf(script_file,
        "\n"
        "\n"
        "################################################################################\n"
        "# create input file\n"
        "echo \"HELO me\" > ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"MAIL FROM:<${FROM_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"RCPT TO:<${TO_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"DATA\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"To: <${REAL_ADDRESS}>\\r\\nFrom: <${FROM_ADDRESS}>\\r\\nSubject: ${TEST_NUM}\\r\\n\\r\\nFOO!\\r\\n.\\r\\n\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"QUIT\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "\n"
        );

      if (spamdyke)
        fprintf(script_file,
          "################################################################################\n"
          "# run test\n"
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "sync\n"
          "\n"
          "output=`grep \"ALLOWED from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ ! -z \"${output}\" ]\n"
          "then\n"
          );
      else
        fprintf(script_file,
          "################################################################################\n"
          "# run test\n"
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "sync\n"
          "\n"
          );

      if (accept_type == ACCEPT_DELIVER)
        fprintf(script_file,
          "  unset outcome\n"
          "  output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "  if [ ! -z \"${output}\" ]\n"
          "  then\n"
          "    if [ ! -z \"${GREP_TARGET}\" ]\n"
          "    then\n"
          "      echo \"Waiting to see if qmail delivers the message...\"\n"
          "      svc -a /service/qmail-send\n"
          "\n"
          "      for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "      do\n"
          "        sync\n"
          "        if [ -e \"${GREP_TARGET}\" ]\n"
          "        then\n"
          "          output=`grep -r \"${FROM_ADDRESS}\" ${GREP_TARGET}`\n"
          "        else\n"
          "          output=\"\"\n"
          "        fi\n"
          "        if [ ! -z \"${output}\" ]\n"
          "        then\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        fi\n"
          "        ${SLEEP_PATH} 1\n"
          "      done\n"
          "\n"
          "      if [ -z \"${outcome}\" ]\n"
          "      then\n"
          "        echo \"DID NOT FIND MESSAGE FROM ${FROM_ADDRESS} IN ${GREP_TARGET}; SHOULD HAVE BEEN DELIVERED\"\n"
          "        outcome=\"failure\"\n"
          "      fi\n"
          "    else\n"
          "      echo SCRIPT BUG: GREP_TARGET IS NOT SET!\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n");
      else if (accept_type == ACCEPT_BOUNCE)
        fprintf(script_file,
          "unset outcome\n"
          "unset MAILDIR\n"
          "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ ! -z \"${output}\" ]\n"
          "then\n"
          "  echo \"Waiting to see if qmail queues or bounces the message...\"\n"
          "  svc -a /service/qmail-send\n"
          "\n"
          "  for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "  do\n"
          "    sync\n"
          "    if [ -f ~postmaster/.qmail ]\n"
          "    then\n"
          "      MAILDIR=`cat ~postmaster/.qmail`\n"
          "      if [ \"${MAILDIR:0:1}\" != \"/\" ]\n"
          "      then\n"
          "        MAILDIR=~postmaster/${MAILDIR}\n"
          "      fi\n"
          "      output=`grep -r ${FROM_ADDRESS} ${MAILDIR}`\n"
          "      if [ ! -z \"${output}\" ]\n"
          "      then\n"
          "        echo \"Found bounced message in ${MAILDIR}\"\n"
          "        outcome=\"success\"\n"
          "        break\n"
          "      fi\n"
          "    fi\n"
          "    if [ -f /var/log/qmail/current ]\n"
          "    then\n"
          "      output=`grep -E \"starting delivery .: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
          "      if [ ! -z \"${output}\" ]\n"
          "      then\n"
          "        echo \"Found log of bounced message in /var/log/qmail/current\"\n"
          "        outcome=\"success\"\n"
          "        break\n"
          "      else\n"
          "        output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
          "        if [ ! -z \"${output}\" ]\n"
          "        then\n"
          "          echo \"Found queued message in /var/qmail/queue/local\"\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        else\n"
          "          output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
          "          if [ ! -z \"${output}\" ]\n"
          "          then\n"
          "            echo \"Found queued message in /var/qmail/queue/remote\"\n"
          "            outcome=\"success\"\n"
          "            break\n"
          "          fi\n"
          "        fi\n"
          "      fi\n"
          "    fi\n"
          "    ${SLEEP_PATH} 1\n"
          "  done\n"
          "\n"
          "  if [ -z \"${outcome}\" ]\n"
          "  then\n"
          "    if [ ! -z \"${MAILDIR}\" ]\n"
          "    then\n"
          "      echo \"COULD NOT FIND BOUNCED MESSAGE FROM ${FROM_ADDRESS} IN ${MAILDIR}; SHOULD HAVE BEEN BOUNCED\"\n"
          "    fi\n"
          "\n"
          "    echo \"COULD NOT FIND QUEUED MESSAGE TO ${REAL_ADDRESS} IN /var/qmail/queue/local OR TO ${FROM_ADDRESS} IN LOGS OR /var/qmail/queue/remote; SHOULD HAVE BEEN QUEUED OR BOUNCED\"\n"
          "    outcome=\"failure\"\n"
          "  fi\n"
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );
      else if (accept_type == ACCEPT_DISCARD)
        fprintf(script_file,
          "unset outcome\n"
          "unset MAILDIR\n"
          "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ ! -z \"${output}\" ]\n"
          "then\n"
          "  echo \"Waiting to see if qmail discards the message...\"\n"
          "  svc -a /service/qmail-send\n"
          "  sync\n"
          "\n"
          "  ${SLEEP_PATH} 5\n"
          "  if [ -f ~postmaster/.qmail ]\n"
          "  then\n"
          "    MAILDIR=`cat ~postmaster/.qmail`\n"
          "    if [ \"${MAILDIR:0:1}\" != \"/\" ]\n"
          "    then\n"
          "      MAILDIR=~postmaster/${MAILDIR}\n"
          "    fi\n"
          "    output=`grep -r ${FROM_ADDRESS} ${MAILDIR}`\n"
          "    if [ ! -z \"${output}\" ]\n"
          "    then\n"
          "      echo \"FOUND BOUNCED MESSAGE IN ${MAILDIR} - SHOULD HAVE BEEN DISCARDED\"\n"
          "      outcome=\"failure\"\n"
          "      break\n"
          "    fi\n"
          "  fi\n"
          "  if [ -f /var/log/qmail/current ]\n"
          "  then\n"
          "    output=`grep -E \"starting delivery .: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
          "    if [ ! -z \"${output}\" ]\n"
          "    then\n"
          "      echo \"FOUND LOG OF BOUNCED MESSAGE IN /var/log/qmail/current - SHOULD HAVE BEEN DISCARDED\"\n"
          "      outcome=\"failure\"\n"
          "      break\n"
          "    else\n"
          "      output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
          "      if [ ! -z \"${output}\" ]\n"
          "      then\n"
          "        echo \"FOUND QUEUED MESSAGE IN /var/qmail/queue/local - SHOULD HAVE BEEN DISCARDED\"\n"
          "        outcome=\"failure\"\n"
          "        break\n"
          "      else\n"
          "        output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
          "        if [ ! -z \"${output}\" ]\n"
          "        then\n"
          "          echo \"FOUND QUEUED MESSAGE IN /var/qmail/queue/remote - SHOULD HAVE BEEN DISCARDED\"\n"
          "          outcome=\"failure\"\n"
          "          break\n"
          "        fi\n"
          "      fi\n"
          "    fi\n"
          "  fi\n"
          "\n"
          "  if [ -z \"${outcome}\" ]\n"
          "  then\n"
          "    echo \"Could not find message - must have been discarded\"\n"
          "    outcome=\"success\"\n"
          "  fi\n"
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );
      else if (accept_type == ACCEPT_REMOTE)
        fprintf(script_file,
          "  unset outcome\n"
          "  output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "  if [ ! -z \"${output}\" ]\n"
          "  then\n"
          "    echo \"Waiting to see if qmail delivers the message...\"\n"
          "    svc -a /service/qmail-send\n"
          "\n"
          "    for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "    do\n"
          "      sync\n"
          "      output=`grep -r \"${REAL_ADDRESS}\" /var/qmail/queue/remote/*`\n"
          "      if [ ! -z \"${output}\" ]\n"
          "      then\n"
          "        outcome=\"success\"\n"
          "        break\n"
          "      else\n"
          "        output=`grep -r \"${FROM_ADDRESS}\" /var/qmail/queue/remote/*`\n"
          "        if [ ! -z \"${output}\" ]\n"
          "        then\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        fi\n"
          "      fi\n"
          "      ${SLEEP_PATH} 1\n"
          "    done\n"
          "\n"
          "    if [ -z \"${outcome}\" ]\n"
          "    then\n"
          "      echo \"COULD NOT FIND MESSAGE TO ${REAL_ADDRESS} OR FROM ${FROM_ADDRESS} IN /var/qmail/queue/remote; SHOULD HAVE BEEN DELIVERED REMOTELY\"\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n");

      if (spamdyke)
        fprintf(script_file,
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );

      fprintf(script_file, "\n%s", post);

      if (restart_needed > 0)
        fprintf(script_file,
          "\n"
          "echo Restarting qmail...\n"
          "svc -d /service/qmail-send\n"
          "svc -d /service/qmail-smtpd\n"
          "for i in 2 1 0\n"
          "do\n"
          "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
          "  if [ -z \"${output}\" ]\n"
          "  then\n"
          "    break\n"
          "  else\n"
          "    ${SLEEP_PATH} 1\n"
          "  fi\n"
          "done\n"
          "if [ ! -z \"${output}\" ]\n"
          "then\n"
          "  svc -k /service/qmail-send\n"
          "  svc -k /service/qmail-smtpd\n"
          "fi\n"
          "svc -u /service/qmail-send\n"
          "svc -u /service/qmail-smtpd\n");

      fprintf(script_file, "\n");

      fclose(script_file);
      }

    next_number++;
    }
  else
    {
    if ((total_number % 1000) == 0)
      total_number++;

    total_number++;
    max_strlen = MAXVAL(max_strlen, MAXVAL(strlen_pre, strlen_post));
    }

  return;
  }


void f37__more_lines_in_file();
void f29__does_passwd_contain_temp_username();
void f24__does_wildcard_filename_exist();
void f07__does_recipient_contain_percent();


DECISION(f47__is_forward_address_valid,
  "forward address is valid",
  "GREP_TARGET=${VALID_GREP_TARGET}\n"
  "echo \"&${VALID_ADDRESS}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  accept_message(ACCEPT_DELIVER);
  lines_in_file--;,
  "",
  "forward address is not valid",
  "REJECT_GREP_TARGET=${VALID_GREP_TARGET}\n"
  "echo \"&${INVALID_ADDRESS}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f37__more_lines_in_file();
  lines_in_file--;,
  "")


DECISION_LIMITED(f46__is_forward_address_same_as_current,
  1,
  "forward address is the same as the current address",
  "GREP_TARGET=${VALID_GREP_TARGET}\n"
  "echo \"&${TO_ADDRESS}${REAL_RELAYCLIENT}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f37__more_lines_in_file();
  lines_in_file--;,
  "",
  -1,
  "forward address is not the same as the current address",
  "",
  f47__is_forward_address_valid();,
  "")


DECISION(f45__is_line_forward_address,
  "line appears to be a forward address",
  "",
  f46__is_forward_address_same_as_current();,
  "",
  "line does not appear to be a forward address",
  "REJECT_GREP_TARGET=${VALID_GREP_TARGET}\n"
  "echo \"-${VALID_ADDRESS}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f37__more_lines_in_file();
  lines_in_file--;,
  "")


DECISION_LIMITED(f44__does_program_exist,
  -1,
  "program exists",
  "GREP_TARGET=${HOMEDIR}/mail_output\n"
  "cp ${TOFILE_PATH} ${TMPDIR}/${TEST_NUM}-f43_tofile.${LEVEL_f43__is_line_program}\n"
  "chmod 755 ${TMPDIR}/${TEST_NUM}-f43_tofile.${LEVEL_f43__is_line_program}\n",
  if (!forwards_only)
    accept_message(ACCEPT_DELIVER);,
  "",
  1,
  "program does not exist",
  "REJECT_GREP_TARGET=${HOMEDIR}/mail_output\n"
  "rm -f ${TMPDIR}/${TEST_NUM}-f43_tofile.${LEVEL_f43__is_line_program}\n",
  reject_message((forwards_only) ? REJECTION_INVALID : REJECTION_UNAVAILABLE);,
  "")


DECISION(f43__is_line_program,
  "line in file is a program",
  "echo \"|${TMPDIR}/${TEST_NUM}-f43_tofile.${LEVEL_f43__is_line_program} -f ${HOMEDIR}/mail_output -t 60\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f44__does_program_exist();
  lines_in_file--;,
  "",
  "line in file is not a program, it must be a forward address",
  "",
  f45__is_line_forward_address();,
  "")


DECISION_LIMITED(f42__does_maildir_exist,
  -1,
  "maildir exists",
  "GREP_TARGET=${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}/new\n"
  "mkdir -p ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}\n"
  "mkdir ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}/cur\n"
  "mkdir ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}/new\n"
  "mkdir ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}/tmp\n"
  "sync\n"
  "chmod -R 700 ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}\n"
  "chown -R $6 ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}\n"
  "chgrp -R $7 ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}\n",
  if (!forwards_only)
    accept_message(ACCEPT_DELIVER);,
  "",
  1,
  "maildir does not exist",
  "REJECT_GREP_TARGET=${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}/new\n"
  "rm -rf ${HOMEDIR}/maildir.${LEVEL_f41__is_line_maildir}\n",
  reject_message((forwards_only) ? REJECTION_INVALID : REJECTION_UNAVAILABLE);,
  "")


DECISION(f41__is_line_maildir,
  "line in file is a maildir",
  "echo \"./maildir.${LEVEL_f41__is_line_maildir}/\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f42__does_maildir_exist();
  lines_in_file--;,
  "",
  "line in file is not a maildir",
  "",
  f43__is_line_program();,
  "")


DECISION_LIMITED(f40__is_mbox_writeable,
  -1,
  "mbox is writeable",
  "GREP_TARGET=${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "touch ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "sync\n"
  "chmod 600 ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "chown $6 ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "chgrp $7 ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n",
  if (!forwards_only)
    accept_message(ACCEPT_DELIVER);,
  "",
  1,
  "mbox is not writeable",
  "REJECT_GREP_TARGET=${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "touch ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "sync\n"
  "chown root ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n"
  "chmod 600 ${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\n",
  reject_message((forwards_only) ? REJECTION_INVALID : REJECTION_UNAVAILABLE);,
  "")


DECISION(f39__is_line_mbox,
  "line in file is an mbox",
  "echo \"${HOMEDIR}/mbox.${LEVEL_f39__is_line_mbox}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f40__is_mbox_writeable();
  lines_in_file--;,
  "",
  "line in file is not an mbox",
  "",
  f41__is_line_maildir();,
  "")


DECISION_LIMITED(f38__is_line_comment,
  1,
  "line in file is a comment",
  "REJECT_GREP_TARGET=${HOMEDIR}/`cat /var/qmail/control/defaultdelivery`\n"
  "echo \"# This line is a meaningless comment.\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  comments_in_file++;
  if (!using_defaultdelivery)
    f37__more_lines_in_file();
  comments_in_file--;
  lines_in_file--;,
  "",
  -1,
  "line in file is not a comment",
  "",
  f39__is_line_mbox();,
  "")


DECISION_LIMITED(f37__more_lines_in_file,
  2,
  "more lines in file",
  "",
  if (!using_defaultdelivery ||
      (reentered_y == 1))
    f38__is_line_comment();,
  "",
  -1,
  "no more lines in file",
  "",
  if (lines_in_file > 0)
    {
    if ((comments_in_file > 0) &&
        !spamdyke)
      accept_message(ACCEPT_DISCARD);
    else
      reject_message(REJECTION_INVALID);
    },
  "")


DECISION(f36__is_first_line_blank,
  "first line is blank",
  "echo \"\" > ${TEMP_FILENAME}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "",
  "first line is not blank",
  "",
  f37__more_lines_in_file();,
  "")


DECISION(f35__is_file_owner_executable,
  "file is owner-executable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_EXECUTABLE_MODE[${LEVEL_f35__is_file_owner_executable}]=${st_mode}\n"
  "chmod u+x ${TEMP_FILENAME}\n",
  forwards_only = 1;
  f36__is_first_line_blank();
  forwards_only = 0;,
  "chmod ${FILE_EXECUTABLE_MODE[${LEVEL_f35__is_file_owner_executable}]} ${TEMP_FILENAME}\n",
  "file is not owner-executable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_EXECUTABLE_MODE[${LEVEL_f35__is_file_owner_executable}]=${st_mode}\n"
  "chmod u-x ${TEMP_FILENAME}\n",
  f36__is_first_line_blank();,
  "chmod ${FILE_EXECUTABLE_MODE[${LEVEL_f35__is_file_owner_executable}]} ${TEMP_FILENAME}\n")


DECISION(f34__is_file_empty,
  "file is empty",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_EMPTY_MODE[${LEVEL_f34__is_file_empty}]=${st_mode}\n"
  "rm -f ${TEMP_FILENAME}\n"
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n"
  "chmod ${FILE_EMPTY_MODE[${LEVEL_f34__is_file_empty}]} ${TEMP_FILENAME}\n"
  "FILE_EMPTY_TEMP_FILENAME[${LEVEL_f34__is_file_empty}]=${TEMP_FILENAME}\n"
  "TEMP_FILENAME=/var/qmail/control/defaultdelivery\n"
  "if [ -f /var/qmail/control/defaultdelivery ]\n"
  "then\n"
  "  cp /var/qmail/control/defaultdelivery ${TMPDIR}/${TEST_NUM}-f34_defaultdelivery.bu.${LEVEL_f34__is_file_empty}\n"
  "  rm -f /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "touch /var/qmail/control/defaultdelivery\n",
  using_defaultdelivery = 1;
  restart_needed++;
  f37__more_lines_in_file();
  restart_needed--;
  using_defaultdelivery = 0;,
  "TEMP_FILENAME=${FILE_EMPTY_TEMP_FILENAME[${LEVEL_f34__is_file_empty}]}\n"
  "rm -f /var/qmail/control/defaultdelivery\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f34_defaultdelivery.bu.${LEVEL_f34__is_file_empty} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_defaultdelivery.bu.${LEVEL_f34__is_file_empty} /var/qmail/control/defaultdelivery\n"
  "fi\n",
  "file is not empty",
  "",
  f35__is_file_owner_executable();,
  "")


DECISION(f33__is_file_world_writable,
  "file is world-writeable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_WRITABLE_MODE[${LEVEL_f33__is_file_world_writable}]=${st_mode}\n"
  "chmod o+w ${TEMP_FILENAME}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "chmod ${FILE_WRITABLE_MODE[${LEVEL_f33__is_file_world_writable}]} ${TEMP_FILENAME}\n",
  "file is not world-writeable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_WRITABLE_MODE[${LEVEL_f33__is_file_world_writable}]=${st_mode}\n"
  "chmod o-w ${TEMP_FILENAME}\n",
  f34__is_file_empty();,
  "chmod ${FILE_WRITABLE_MODE[${LEVEL_f33__is_file_world_writable}]} ${TEMP_FILENAME}\n")


DECISION(f32__is_file_readable,
  "file is readable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_READABLE_MODE[${LEVEL_f32_is_file_readable}]=${st_mode}\n"
  "chmod o+r ${TEMP_FILENAME}\n",
  f33__is_file_world_writable();,
  "chmod ${FILE_READABLE_MODE[${LEVEL_f32_is_file_readable}]} ${TEMP_FILENAME}\n",
  "file is not readable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_READABLE_MODE[${LEVEL_f32_is_file_readable}]=${st_mode}\n"
  "chown root ${TEMP_FILENAME}\n"
  "chmod 600 ${TEMP_FILENAME}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "chmod ${FILE_READABLE_MODE[${LEVEL_f32_is_file_readable}]} ${TEMP_FILENAME}\n")


DECISION(f31__does_passwd_contain_alias,
  "passwd contains alias",
  "HOMEDIR=\"${TMPDIR}/${TEST_NUM}-home.${LEVEL_f31__does_passwd_contain_alias}\"\n"
  "DASH=\"-\"\n"
  "EXT=`echo ${REAL_USERNAME} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f32_master.passwd.bu.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cat /etc/master.passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f32_master.passwd.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  echo \"alias:*************:$6:$7::0:0:alias:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f32_master.passwd.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_master.passwd.${LEVEL_f31__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f32_passwd.bu.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cat /etc/passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f32_passwd.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  echo \"alias:x:$6:$7:alias:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f32_passwd.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_passwd.${LEVEL_f31__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n"
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chown $6 ${HOMEDIR}\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  dash_emptyable = 0;
  dash_empty = 0;
  restart_needed++;
  f24__does_wildcard_filename_exist();
  restart_needed--;
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_master.passwd.bu.${LEVEL_f31__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_passwd.bu.${LEVEL_f31__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n",
  "passwd does not contain alias",
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f32_master.passwd.bu.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cat /etc/master.passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f32_master.passwd.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_master.passwd.${LEVEL_f31__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f32_passwd.bu.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cat /etc/passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f32_passwd.${LEVEL_f31__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_passwd.${LEVEL_f31__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n",
  restart_needed++;
  reject_message(REJECTION_INVALID);
  restart_needed--;,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_master.passwd.bu.${LEVEL_f31__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f32_passwd.bu.${LEVEL_f31__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n")


DECISION_LIMITED(f30__does_temp_username_contain_hyphen,
  2,
  "temp username contains hyphen",
  "USERNAME_HYPHEN_TEMP_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]=${TEMP_USERNAME}\n"
  "TEMP_USERNAME=`echo ${TEMP_USERNAME} | sed -e \"s/-[^-]*$//\"`\n",
  if (num_username_hyphens > 0)
    {
    num_username_hyphens--;
    num_ext_hyphens++;
    username_truncated++;
    f29__does_passwd_contain_temp_username();
    username_truncated--;
    num_ext_hyphens--;
    num_username_hyphens++;
    },
  "TEMP_USERNAME=${USERNAME_HYPHEN_TEMP_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]}\n",
  -1,
  "temp username does not contain a hyphen",
  "USERNAME_HYPHEN_REAL_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]=${REAL_USERNAME}\n"
  "USERNAME_HYPHEN_REAL_ADDRESS[${LEVEL_f30__does_temp_username_contain_hyphen}]=${REAL_ADDRESS}\n"
  "USERNAME_HYPHEN_TO_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]=${TO_USERNAME}\n"
  "USERNAME_HYPHEN_TO_ADDRESS[${LEVEL_f30__does_temp_username_contain_hyphen}]=${TO_ADDRESS}\n"
  "USERNAME_HYPHEN_TEMP_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]=${TEMP_USERNAME}\n"
  "TRUNCATED=${REAL_USERNAME:${#TEMP_USERNAME}:$[${#REAL_USERNAME}-${#TEMP_USERNAME}]}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}`echo ${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#TRUNCATED}]]} | sed -e \"s/-/_/g\"`${TRUNCATED}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=`echo ${TO_USERNAME:0:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-${#TRUNCATED}]} | sed -e \"s/-/_/g\"`${TO_USERNAME:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-${#TRUNCATED}]:${#TRUNCATED}}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}`echo ${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#TRUNCATED}]]} | sed -e \"s/-/_/g\"`${TRUNCATED}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=`echo ${TO_USERNAME:0:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-$[${#TRUNCATED}-${#REAL_RELAYCLIENT}]]} | sed -e \"s/-/_/g\"`${TO_USERNAME:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-$[${#TRUNCATED}-${#REAL_RELAYCLIENT}]]:$[${#TRUNCATED}-${#REAL_RELAYCLIENT}]}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "TEMP_USERNAME=`echo ${TEMP_USERNAME} | sed -e \"s/-/_/g\"`\n",
  f31__does_passwd_contain_alias();,
  "REAL_USERNAME=${USERNAME_HYPHEN_REAL_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]}\n"
  "REAL_ADDRESS=${USERNAME_HYPHEN_REAL_ADDRESS[${LEVEL_f30__does_temp_username_contain_hyphen}]}\n"
  "TO_USERNAME=${USERNAME_HYPHEN_TO_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]}\n"
  "TO_ADDRESS=${USERNAME_HYPHEN_TO_ADDRESS[${LEVEL_f30__does_temp_username_contain_hyphen}]}\n"
  "TEMP_USERNAME=${USERNAME_HYPHEN_TEMP_USERNAME[${LEVEL_f30__does_temp_username_contain_hyphen}]}\n")


DECISION(f29__does_passwd_contain_temp_username,
  "passwd contains temp username",
  "HOMEDIR=\"${TMPDIR}/${TEST_NUM}-home.${LEVEL_f29__does_passwd_contain_temp_username}\"\n"
  "if [ \"${TEMP_USERNAME}\" == \"${REAL_USERNAME}\" ]\n"
  "then\n"
  "  DASH=\"\"\n"
  "else\n"
  "  DASH=\"-\"\n"
  "fi\n"
  "EXT=`echo ${REAL_USERNAME:${#TEMP_USERNAME}:$[${#REAL_USERNAME}-${#TEMP_USERNAME}]} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\" -e \"s/^-//\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f30_master.passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cat /etc/master.passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f30_master.passwd.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  echo \"${TEMP_USERNAME}:*************:$6:$7::0:0:${TEMP_USERNAME}:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f30_master.passwd.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_master.passwd.${LEVEL_f29__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f30_passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cat /etc/passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f30_passwd.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  echo \"${TEMP_USERNAME}:x:$6:$7:${TEMP_USERNAME}:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f30_passwd.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_passwd.${LEVEL_f29__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n"
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chown $6 ${HOMEDIR}\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  dash_emptyable = 0;
  dash_empty = (username_truncated) ? 0 : 1;
  f24__does_wildcard_filename_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_master.passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n",
  "passwd does not contain temp username",
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f30_master.passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cat /etc/master.passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f30_master.passwd.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_master.passwd.${LEVEL_f29__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f30_passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cat /etc/passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f30_passwd.${LEVEL_f29__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_passwd.${LEVEL_f29__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n",
  f30__does_temp_username_contain_hyphen();,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_master.passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_passwd.bu.${LEVEL_f29__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n")


DECISION(f28__is_temp_username_over_31_chars,
  "temp username is over 31 characters",
  "OVER32_REAL_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]=${REAL_USERNAME}\n"
  "OVER32_TO_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]=${TO_USERNAME}\n"
  "WILDCARD_DOT=\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\n"
  "WILDCARD_DASH=\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}-${WILDCARD_DOT}.${WILDCARD_DASH}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME}-${WILDCARD_DOT}.${WILDCARD_DASH}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}-${WILDCARD_DOT}.${WILDCARD_DASH}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "TEMP_USERNAME=${REAL_USERNAME:0:32}\n",
  num_username_hyphens++;
  username_truncated++;
  f30__does_temp_username_contain_hyphen();
  username_truncated--;,
  "REAL_USERNAME=${OVER32_REAL_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${OVER32_TO_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n",
  "temp username is not over 31 characters",
  "OVER32_REAL_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]=${REAL_USERNAME}\n"
  "OVER32_TO_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]=${TO_USERNAME}\n"
  "WILDCARD_DOT=\0\n"
  "WILDCARD_DASH=\0\n"
  "WILDCARD_DASH2=\0\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#REAL_RELAYCLIENT}]]}\n"
  "  if [ ${#REAL_USERNAME} -gt $[25-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#REAL_RELAYCLIENT}]] ]\n"
  "  then\n"
  "    STRLEN_REAL_USERNAME=$[25-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#REAL_RELAYCLIENT}]]\n"
  "  else\n"
  "    STRLEN_REAL_USERNAME=${#REAL_USERNAME}\n"
  "  fi\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-${#VIRTUALDOMAINS_USERNAME_ADDITION}]}\n"
  "  if [ ${#REAL_USERNAME} -gt $[25-${#VIRTUALDOMAINS_USERNAME_ADDITION}] ]\n"
  "  then\n"
  "    STRLEN_REAL_USERNAME=$[25-${#VIRTUALDOMAINS_USERNAME_ADDITION}]\n"
  "  else\n"
  "    STRLEN_REAL_USERNAME=${#REAL_USERNAME}\n"
  "  fi\n"
  "fi\n"
  "REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}`echo ${REAL_USERNAME:0:${STRLEN_REAL_USERNAME}} | sed -e \"s/-*$//\"`.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}\n"
  "TO_USERNAME=`echo ${TO_USERNAME:0:${STRLEN_REAL_USERNAME}} | sed -e \"s/-*$//\"`.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "  REAL_USERNAME=${REAL_USERNAME}${REAL_RELAYCLIENT}\n"
  "fi\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TEMP_USERNAME=${REAL_USERNAME}\n",
  num_username_hyphens = 2;
  f29__does_passwd_contain_temp_username();,
  "REAL_USERNAME=${OVER32_REAL_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${OVER32_TO_USERNAME[${LEVEL_f28__is_temp_username_over_31_chars}]}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n")


DECISION(f27__is_wildcard_dash_empty,
  "dash field is empty",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f28_assign.bu.${LEVEL_f27__is_wildcard_dash_empty}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f28_assign.bu.${LEVEL_f27__is_wildcard_dash_empty}`\n"
  "do\n"
  "  if [ \"${line:0:1}\" == \"=\" ]\n"
  "  then\n"
  "    pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "    if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "    then\n"
  "      echo ${line} >> ${TMPDIR}/${TEST_NUM}-f28_assign.${LEVEL_f27__is_wildcard_dash_empty}\n"
  "    fi\n"
  "  else\n"
  "    if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "      pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "      if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "      then\n"
  "        if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "        then\n"
  "          echo ${line} >> ${TMPDIR}/${TEST_NUM}-f28_assign.${LEVEL_f27__is_wildcard_dash_empty}\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "DASH=\"\"\n"
  "TEMP_FILENAME=/var/qmail/control/defaultdelivery\n"
  "if [ -f /var/qmail/control/defaultdelivery ]\n"
  "then\n"
  "  cp /var/qmail/control/defaultdelivery ${TMPDIR}/${TEST_NUM}-f28_defaultdelivery.bu.${LEVEL_f27__is_wildcard_dash_empty}\n"
  "  rm -f /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "touch /var/qmail/control/defaultdelivery\n"
  "echo \"+${PATTERN}:${USER}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f28_assign.${LEVEL_f27__is_wildcard_dash_empty}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f28_assign.${LEVEL_f27__is_wildcard_dash_empty}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f28_assign.${LEVEL_f27__is_wildcard_dash_empty} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  if (dash_emptyable)
    {
    using_defaultdelivery = 1;
    restart_needed++;
    f37__more_lines_in_file();
    restart_needed--;
    using_defaultdelivery = 0;
    },
  "rm -f /var/qmail/control/defaultdelivery\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f28_defaultdelivery.bu.${LEVEL_f27__is_wildcard_dash_empty} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f28_defaultdelivery.bu.${LEVEL_f27__is_wildcard_dash_empty} /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "cp ${TMPDIR}/${TEST_NUM}-f28_assign.bu.${LEVEL_f27__is_wildcard_dash_empty} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "dash field is not empty",
  "",
  if (!dash_empty)
    reject_message(REJECTION_INVALID);,
  "")


DECISION(f26__is_ext_empty,
  "ext is empty",
  "",
  if (num_ext_hyphens < 0)
    f27__is_wildcard_dash_empty();,
  "",
  "ext is not empty",
  "EXT_EMPTY_EXT[${LEVEL_f26__is_ext_empty}]=${EXT}\n"
  "EXT_EMPTY_TEMP_FILENAME[${LEVEL_f26__is_ext_empty}]=${TEMP_FILENAME}\n"
  "EXT=\"\"\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}default\n",
  if (num_ext_hyphens == 0)
    {
    num_ext_hyphens--;
    f24__does_wildcard_filename_exist();
    num_ext_hyphens++;
    },
  "EXT=${EXT_EMPTY_EXT[${LEVEL_f26__is_ext_empty}]}\n"
  "TEMP_FILENAME=${EXT_EMPTY_TEMP_FILENAME[${LEVEL_f26__is_ext_empty}]}\n")


DECISION_LIMITED(f25__does_ext_contain_hyphen,
  1,
  "ext contains hyphen",
  "EXT_HYPHEN_EXT[${LEVEL_f25__does_ext_contain_hyphen}]=${EXT}\n"
  "EXT_HYPHEN_TEMP_FILENAME[${LEVEL_f25__does_ext_contain_hyphen}]=${TEMP_FILENAME}\n"
  "EXT=`echo ${EXT} | sed -e \"s/-[^-]*$//\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}-default\n",
  if (num_ext_hyphens > 0)
    {
    num_ext_hyphens--;
    f24__does_wildcard_filename_exist();
    num_ext_hyphens++;
    },
  "EXT=${EXT_HYPHEN_EXT[${LEVEL_f25__does_ext_contain_hyphen}]}\n"
  "TEMP_FILENAME=${EXT_HYPHEN_TEMP_FILENAME[${LEVEL_f25__does_ext_contain_hyphen}]}\n",
  -1,
  "ext does not contain a hyphen",
  "",
  if (num_ext_hyphens <= 0)
    f26__is_ext_empty();,
  "")


DECISION(f24__does_wildcard_filename_exist,
  "constructed filename exists",
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n",
  f32__is_file_readable();,
  "",
  "constructed filename does not exist",
  "rm -f ${TEMP_FILENAME}\n",
  f25__does_ext_contain_hyphen();,
  "")


DECISION(f23__is_wildcard_home_directory_sticky,
  "home directory is sticky",
  "chmod +t ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "chmod -t ${HOMEDIR}\n",
  "home directory is not sticky",
  "chmod -t ${HOMEDIR}\n",
  f24__does_wildcard_filename_exist();,
  "")


DECISION(f22__is_wildcard_home_directory_world_writable,
  "home directory is world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f22__is_wildcard_home_directory_world_writable}]=${st_mode}\n"
  "chmod o+w ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "chmod ${WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f22__is_wildcard_home_directory_world_writable}]} ${HOMEDIR}\n",
  "home directory is not world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f22__is_wildcard_home_directory_world_writable}]=${st_mode}\n"
  "chmod o-w ${HOMEDIR}\n",
  f23__is_wildcard_home_directory_sticky();,
  "chmod ${WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f22__is_wildcard_home_directory_world_writable}]} ${HOMEDIR}\n")


DECISION(f21__is_wildcard_home_directory_readable,
  "home directory is readable",
  "chmod u+r ${HOMEDIR}\n",
  f22__is_wildcard_home_directory_world_writable();,
  "",
  "home directory is not readable",
  "chmod u-r,g-r,o-r ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "")


DECISION(f20__does_wildcard_home_directory_exist,
  "home directory exists",
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chmod 700 ${HOMEDIR}\n"
  "chown -R $6 ${HOMEDIR}\n"
  "chgrp -R $7 ${HOMEDIR}\n",
  f21__is_wildcard_home_directory_readable();,
  "",
  "home directory does not exist",
  "rm -rf ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "")


DECISION(f19__is_exact_match_dash_empty,
  "dash field is empty",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f20_assign.bu.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f20_assign.bu.${LEVEL_f19__is_exact_match_dash_empty} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f20_assign.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "for line in `cat /var/qmail/users/assign`\n"
  "do\n"
  "  if [ \"${line:0:1}\" == \"=\" ]\n"
  "  then\n"
  "    pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "    if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "    then\n"
  "      echo ${line} >> ${TMPDIR}/${TEST_NUM}-f20_assign.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "    fi\n"
  "  else\n"
  "    if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "      pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "      if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "      then\n"
  "        if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "        then\n"
  "          echo ${line} >> ${TMPDIR}/${TEST_NUM}-f20_assign.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "DASH=\"\"\n"
  "TEMP_FILENAME=/var/qmail/control/defaultdelivery\n"
  "if [ -f /var/qmail/control/defaultdelivery ]\n"
  "then\n"
  "  cp /var/qmail/control/defaultdelivery ${TMPDIR}/${TEST_NUM}-f20_defaultdelivery.bu.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "  rm /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "touch /var/qmail/control/defaultdelivery\n"
  "echo \"=${PATTERN}:${REAL_USERNAME}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f20_assign.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f20_assign.${LEVEL_f19__is_exact_match_dash_empty}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f20_assign.${LEVEL_f19__is_exact_match_dash_empty} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  using_defaultdelivery = 1;
  restart_needed++;
  f37__more_lines_in_file();
  restart_needed--;
  using_defaultdelivery = 0;,
  "cp ${TMPDIR}/${TEST_NUM}-f20_assign.bu.${LEVEL_f19__is_exact_match_dash_empty} /var/qmail/users/assign\n"
  "rm /var/qmail/control/defaultdelivery\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f20_defaultdelivery.bu.${LEVEL_f19__is_exact_match_dash_empty} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f20_defaultdelivery.bu.${LEVEL_f19__is_exact_match_dash_empty} /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "dash field is not empty",
  "",
  if (!dash_empty)
    reject_message(REJECTION_INVALID);,
  "")


DECISION(f18__does_exact_match_filename_exist_default,
  "constructed filename exists",
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n",
  f32__is_file_readable();,
  "",
  "constructed filename does not exist",
  "rm -f ${TEMP_FILENAME}\n",
  f19__is_exact_match_dash_empty();,
  "")


DECISION(f17__does_exact_match_filename_exist,
  "constructed filename exists",
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n",
  f32__is_file_readable();,
  "",
  "constructed filename does not exist",
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}default\n",
  f18__does_exact_match_filename_exist_default();,
  "")


DECISION(f16__is_exact_match_home_directory_sticky,
  "home directory is sticky",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f16__is_exact_match_home_directory_sticky}]=${st_mode}\n"
  "chmod +t ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "chmod ${EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f16__is_exact_match_home_directory_sticky}]} ${HOMEDIR}\n",
  "home directory is not sticky",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f16__is_exact_match_home_directory_sticky}]=${st_mode}\n"
  "chmod -t ${HOMEDIR}\n",
  f17__does_exact_match_filename_exist();,
  "chmod ${EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f16__is_exact_match_home_directory_sticky}]} ${HOMEDIR}\n")


DECISION(f15__is_exact_match_home_directory_world_writable,
  "home directory is world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f15__is_exact_match_home_directory_world_writable}]=${st_mode}\n"
  "chmod o+w ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "chmod ${EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f15__is_exact_match_home_directory_world_writable}]} ${HOMEDIR}\n",
  "home directory is not world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f15__is_exact_match_home_directory_world_writable}]=${st_mode}\n"
  "chmod o-w ${HOMEDIR}\n",
  f16__is_exact_match_home_directory_sticky();,
  "chmod ${EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f15__is_exact_match_home_directory_world_writable}]} ${HOMEDIR}\n")


DECISION(f14__is_exact_match_home_directory_readable,
  "home directory is readable",
  "chmod u+r ${HOMEDIR}\n",
  f15__is_exact_match_home_directory_world_writable();,
  "",
  "home directory is not readable",
  "chmod u-r,g-r,o-r ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "")


DECISION(f13__does_exact_match_home_directory_exist,
  "home directory exists",
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chmod 700 ${HOMEDIR}\n"
  "chown -R $6 ${HOMEDIR}\n"
  "chgrp -R $7 ${HOMEDIR}\n",
  f14__is_exact_match_home_directory_readable();,
  "",
  "home directory does not exist",
  "rm -rf ${HOMEDIR}\n",
  reject_message(REJECTION_UNAVAILABLE);,
  "")


DECISION(f12__does_cdb_contain_wildcard_match,
  "cdb contains recipient wildcard match",
  "WILDCARD_REAL_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]=${REAL_USERNAME}\n"
  "WILDCARD_TO_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]=${TO_USERNAME}\n"
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_wildcard_match}`\n"
  "do\n"
  "  if [ \"${line}\" != \".\" ]\n"
  "  then\n"
  "    if [ \"${line:0:1}\" == \"=\" ]\n"
  "    then\n"
  "      pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "      if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "      then\n"
  "        echo ${line} >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "      fi\n"
  "    else\n"
  "      if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "        pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "        if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "        then\n"
  "          if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "          then\n"
  "            echo ${line} >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "          fi\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "PATTERN=`echo ${REAL_USERNAME} | sed -e \"s/-[^-]*$/-/\"`\n"
  "WILDCARD_ADDITION=\0\0\0\0\0\n"
  "USER=${REAL_DOMAIN}\n"
  "HOMEDIR=${TMPDIR}/${TEST_NUM}-home.${LEVEL_f12__does_cdb_contain_wildcard_match}.\0\0\0\0\0\n"
  "DASH=-DASH${LEVEL_f12__does_cdb_contain_wildcard_match}.\0\0\0\0\0-\n"
  "FILE_EXT=EXT${LEVEL_f12__does_cdb_contain_wildcard_match}.\0\0\0\0\0-\0\0\0\0\0\n"
  "EXT=`echo ${FILE_EXT} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}${REAL_USERNAME:${#PATTERN}:$[${#REAL_USERNAME}-$[${#PATTERN}+${#REAL_RELAYCLIENT}]]}:${WILDCARD_ADDITION}${REAL_RELAYCLIENT}\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}.${WILDCARD_ADDITION}${REAL_RELAYCLIENT}\n"
  "else\n"
  "  TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}${REAL_USERNAME:${#PATTERN}:$[${#REAL_USERNAME}-${#PATTERN}]}:${WILDCARD_ADDITION}\n"
  "  REAL_USERNAME=${REAL_USERNAME}.${WILDCARD_ADDITION}\n"
  "fi\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}.${WILDCARD_ADDITION}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "echo \"+${PATTERN}:${USER}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  num_ext_hyphens = 1;
  dash_emptyable = 1;
  dash_empty = 0;
  f20__does_wildcard_home_directory_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_wildcard_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "REAL_USERNAME=${WILDCARD_REAL_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${WILDCARD_TO_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n",
  "cdb does not contain recipient wildcard match",
  "WILDCARD_REAL_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]=${REAL_USERNAME}\n"
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_wildcard_match}`\n"
  "do\n"
  "  if [ \"${line}\" != \".\" ]\n"
  "  then\n"
  "    if [ \"${line:0:1}\" == \"=\" ]\n"
  "    then\n"
  "      pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "      if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "      then\n"
  "        echo ${line} >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "      fi\n"
  "    else\n"
  "      if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "        pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "        if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "        then\n"
  "          if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "          then\n"
  "            echo ${line} >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "          fi\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_wildcard_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "WILDCARD_DOT=\0\n"
  "WILDCARD_DASH=\0\n"
  "WILDCARD_DASH2=\0\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME}.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "TEMP_USERNAME=${REAL_USERNAME}\n",
  num_username_hyphens = 2;
  num_ext_hyphens = -1;
  f28__is_temp_username_over_31_chars();,
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_wildcard_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "REAL_USERNAME=${WILDCARD_REAL_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${WILDCARD_TO_USERNAME[${LEVEL_f12__does_cdb_contain_wildcard_match}]}\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n")


DECISION(f11__does_cdb_contain_exact_match,
  "cdb contains recipient exact match",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f11_assign.bu.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f11_assign.bu.${LEVEL_f11__does_cdb_contain_exact_match} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "PATTERN=${REAL_USERNAME}\n"
  "USER=${REAL_USERNAME}\n"
  "HOMEDIR=${TMPDIR}/${TEST_NUM}-home.${LEVEL_f11__does_cdb_contain_exact_match}.\0\0\0\0\0\n"
  "DASH=-DASH${LEVEL_f11__does_cdb_contain_exact_match}.\0\0\0\0\0-\n"
  "DASH_EMPTY=\"\"\n"
  "FILE_EXT=EXT${LEVEL_f11__does_cdb_contain_exact_match}.\0\0\0\0\0\n"
  "EXT=`echo ${FILE_EXT} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "echo \"=${PATTERN}:${REAL_USERNAME}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "echo \"=${VALID_USERNAME}:${VALID_USERNAME}:$6:$7:${VALID_HOMEDIR}:::\" >> ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  dash_emptyable = 1;
  dash_empty = 0;
  f13__does_exact_match_home_directory_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "cp ${TMPDIR}/${TEST_NUM}-f11_assign.bu.${LEVEL_f11__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "cdb does not contain recipient exact match",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f11_assign.bu.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f11_assign.bu.${LEVEL_f11__does_cdb_contain_exact_match} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "echo \"=${VALID_USERNAME}:${VALID_USERNAME}:$6:$7:${VALID_HOMEDIR}:::\" >> ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f11_assign.${LEVEL_f11__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  f12__does_cdb_contain_wildcard_match();,
  "cp ${TMPDIR}/${TEST_NUM}-f11_assign.bu.${LEVEL_f11__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n")


DECISION(f10__is_domain_in_virtualdomains,
  "domain is in virtualdomains",
  "VIRTUALDOMAINS_REAL_USERNAME[${LEVEL_f10__is_domain_in_virtualdomains}]=${REAL_USERNAME}\n"
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains}\n"
  "echo \"${REAL_DOMAIN}:${REAL_DOMAIN}-\" > ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.${LEVEL_f10__is_domain_in_virtualdomains}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.${LEVEL_f10__is_domain_in_virtualdomains} /var/qmail/control/virtualdomains\n"
  "VIRTUALDOMAINS_USERNAME_ADDITION=${REAL_DOMAIN}--\n"
  "REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}${REAL_USERNAME}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n",
  restart_needed++;
  f11__does_cdb_contain_exact_match();
  restart_needed--;,
  "REAL_USERNAME=${VIRTUALDOMAINS_REAL_USERNAME[${LEVEL_f10__is_domain_in_virtualdomains}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains} /var/qmail/control/virtualdomains\n",
  "domain is not in virtualdomains",
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains} | sed -e \"/^${REAL_DOMAIN}:/d\" > ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.${LEVEL_f10__is_domain_in_virtualdomains}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.${LEVEL_f10__is_domain_in_virtualdomains} /var/qmail/control/virtualdomains\n",
  restart_needed++;
  accept_message(ACCEPT_REMOTE);
  restart_needed--;,
  "cp ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains} /var/qmail/control/virtualdomains\n")


DECISION(f09__is_domain_in_locals,
  "domain is in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "echo ${TO_DOMAIN}${REAL_RELAYCLIENT} >> ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "echo ${VALID_DOMAIN} >> ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n",
  restart_needed++;
  f11__does_cdb_contain_exact_match();
  restart_needed--;,
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n",
  "domain is not in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals} | sed -e \"/^${REAL_DOMAIN}$/d\" > ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "echo ${VALID_DOMAIN} >> ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n",
  restart_needed++;
  f10__is_domain_in_virtualdomains();
  restart_needed--;,
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n")


DECISION_LIMITED(f08__is_domain_in_percenthack,
  1,
  "domain is in percenthack",
  "PERCENTHACK_TO_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${TO_USERNAME}\n"
  "PERCENTHACK_TO_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${TO_DOMAIN}\n"
  "PERCENTHACK_TO_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${TO_ADDRESS}\n"
  "PERCENTHACK_REAL_RELAYCLIENT[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_RELAYCLIENT}\n"
  "PERCENTHACK_REAL_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_USERNAME}\n"
  "PERCENTHACK_REAL_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_DOMAIN}\n"
  "PERCENTHACK_REAL_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_ADDRESS}\n"
  "PERCENTHACK_SAVED_ADDITION[${LEVEL_f08__is_domain_in_percenthack}]=${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ -f /var/qmail/control/percenthack ]\n"
  "then\n"
  "  cp /var/qmail/control/percenthack ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack} ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "else\n"
  "  touch ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "fi\n"
  "NEW_DOMAIN=\0\0.test\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  PERCENTHACK_USERNAME_ADDITION=${PERCENTHACK_USERNAME_ADDITION}%${NEW_DOMAIN}\n"
  "  echo ${REAL_DOMAIN} >> ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "  TO_USERNAME=${TO_USERNAME}%${NEW_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}\n"
  "  REAL_DOMAIN=${NEW_DOMAIN}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  REAL_RELAYCLIENT=\"\"\n"
  "else\n"
  "  PERCENTHACK_USERNAME_ADDITION=${PERCENTHACK_USERNAME_ADDITION}%${TO_DOMAIN}\n"
  "  REAL_DOMAIN=${TO_DOMAIN}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=${TO_USERNAME}%${TO_DOMAIN}\n"
  "  TO_DOMAIN=${NEW_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "  if [ \"${IN_RCPTHOSTS}\" == \"true\" ]\n"
  "  then\n"
  "    cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.bu.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    echo \"${NEW_DOMAIN}${REAL_RELAYCLIENT}\" >> ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    cp ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/rcpthosts\n"
  "  fi\n"
  "  if [ \"${IN_MORERCPTHOSTS}\" == \"true\" ]\n"
  "  then\n"
  "    cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.bu.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    echo \"${NEW_DOMAIN}${REAL_RELAYCLIENT}\" >> ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    cp ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/morercpthosts\n"
  "    /var/qmail/bin/qmail-newmrh\n"
  "  fi\n"
  "  echo ${NEW_DOMAIN}${REAL_RELAYCLIENT} >> ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "  REAL_RELAYCLIENT=\"\"\n"
  "fi\n"
  "cp ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/percenthack\n",
  restart_needed++;
  f07__does_recipient_contain_percent();
  restart_needed--;,
  "TO_USERNAME=${PERCENTHACK_TO_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "TO_DOMAIN=${PERCENTHACK_TO_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "TO_ADDRESS=${PERCENTHACK_TO_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "REAL_RELAYCLIENT=${PERCENTHACK_REAL_RELAYCLIENT[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "REAL_USERNAME=${PERCENTHACK_REAL_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "REAL_DOMAIN=${PERCENTHACK_REAL_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "REAL_ADDRESS=${PERCENTHACK_REAL_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "PERCENTHACK_USERNAME_ADDITION=${PERCENTHACK_SAVED_ADDITION[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/percenthack\n"
  "else\n"
  "  rm -f /var/qmail/control/percenthack\n"
  "fi\n"
  "if [ ! -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  if [ \"${IN_RCPTHOSTS}\" == \"true\" ]\n"
  "  then\n"
  "    cp ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/rcpthosts\n"
  "  fi\n"
  "  if [ \"${IN_MORERCPTHOSTS}\" == \"true\" ]\n"
  "  then\n"
  "    if [ -f ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} ]\n"
  "    then\n"
  "      cp ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/morercpthosts\n"
  "      /var/qmail/bin/qmail-newmrh\n"
  "    else\n"
  "      rm -f /var/qmail/control/morercpthosts /var/qmail/control/morercpthosts\n"
  "    fi\n"
  "  fi\n"
  "fi\n",
  -1,
  "domain is not in percenthack",
  "PERCENTHACK_TO_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${TO_USERNAME}\n"
  "PERCENTHACK_TO_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${TO_DOMAIN}\n"
  "PERCENTHACK_TO_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${TO_ADDRESS}\n"
  "PERCENTHACK_REAL_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_USERNAME}\n"
  "PERCENTHACK_REAL_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_DOMAIN}\n"
  "PERCENTHACK_REAL_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_ADDRESS}\n"
  "NEW_USERNAME=\"%\0\0\0\0\0\"\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}${NEW_USERNAME}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}${NEW_USERNAME}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME}${NEW_USERNAME}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}${NEW_USERNAME}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "fi\n",
  restart_needed++;
  f09__is_domain_in_locals();
  restart_needed--;,
  "REAL_USERNAME=${PERCENTHACK_REAL_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "REAL_DOMAIN=${PERCENTHACK_REAL_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "REAL_ADDRESS=${PERCENTHACK_REAL_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "TO_USERNAME=${PERCENTHACK_TO_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "TO_DOMAIN=${PERCENTHACK_TO_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]}\n"
  "TO_ADDRESS=${PERCENTHACK_TO_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]}\n")


DECISION(f07__does_recipient_contain_percent,
  "recipient contains a(nother) percent sign",
  "",
  f08__is_domain_in_percenthack();,
  "",
  "recipient does not contain a(nother) percent sign",
  "",
  f09__is_domain_in_locals();,
  "")


DECISION(f06__is_domain_in_morercpthosts,
  "domain is in morercpthosts",
  "IN_MORERCPTHOSTS=\"true\"\n"
  "if [ -f /var/qmail/control/morercpthosts ]\n"
  "then\n"
  "  cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "fi\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "/var/qmail/bin/qmail-newmrh\n",
  f07__does_recipient_contain_percent();,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "else\n"
  "  rm -f /var/qmail/control/morercpthosts /var/qmail/control/morercpthosts.cdb\n"
  "fi\n",
  "domain is not in morercpthosts",
  "IN_MORERCPTHOSTS=\"false\"\n"
  "if [ -f /var/qmail/control/morercpthosts ]\n"
  "then\n"
  "  cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "  cat ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} | sed -e \"/^${REAL_DOMAIN}$/d\" > ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "fi\n",
  reject_message(REJECTION_RELAYING);,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "else\n"
  "  rm -f /var/qmail/control/morercpthosts /var/qmail/control/morercpthosts.cdb\n"
  "fi\n")


DECISION(f05__is_domain_in_rcpthosts,
  "domain is in rcpthosts",
  "IN_RCPTHOSTS=\"true\"\n"
  "cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n",
  f07__does_recipient_contain_percent();,
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n",
  "domain is not in rcpthosts",
  "IN_RCPTHOSTS=\"false\"\n"
  "cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts} | sed -e \"/^${REAL_DOMAIN}$/d\" > ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n",
  f06__is_domain_in_morercpthosts();,
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n")


DECISION(f04__does_me_exist,
  "me exists",
  "ME_REAL_DOMAIN[${LEVEL_f04__does_me_exist}]=${REAL_DOMAIN}\n"
  "if [ -f /var/qmail/control/me ]\n"
  "then\n"
  "  cp /var/qmail/control/me ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist}\n"
  "fi\n"
  "REAL_DOMAIN=\0\0.test\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f04_me.${LEVEL_f04__does_me_exist}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f04_me.${LEVEL_f04__does_me_exist} /var/qmail/control/me\n",
  restart_needed++;
  f07__does_recipient_contain_percent();
  restart_needed--;,
  "REAL_DOMAIN=${ME_REAL_DOMAIN[${LEVEL_f04__does_me_exist}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} /var/qmail/control/me\n"
  "else\n"
  "  rm /var/qmail/control/me\n"
  "fi\n",
  "me does not exist",
  "ME_REAL_DOMAIN[${LEVEL_f04__does_me_exist}]=${REAL_DOMAIN}\n"
  "REAL_DOMAIN=envnoathost\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f /var/qmail/control/me ]\n"
  "then\n"
  "  cp /var/qmail/control/me ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist}\n"
  "  rm /var/qmail/control/me\n"
  "fi\n",
  restart_needed++;
  reject_message(REJECTION_CANNOT_START);
  restart_needed--;,
  "REAL_DOMAIN=${ME_REAL_DOMAIN[${LEVEL_f04__does_me_exist}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} /var/qmail/control/me\n"
  "fi\n")


DECISION(f03__does_envnoathost_exist,
  "envnoathost exists",
  "if [ -f /var/qmail/control/envnoathost ]\n"
  "then\n"
  "  cp /var/qmail/control/envnoathost ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist}\n"
  "fi\n"
  "ENVNOATHOST_REAL_DOMAIN[${LEVEL_f03__does_envnoathost_exist}]=${REAL_DOMAIN}\n"
  "REAL_DOMAIN=\0\0.test\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f03_envnoathost.${LEVEL_f03__does_envnoathost_exist}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f03_envnoathost.${LEVEL_f03__does_envnoathost_exist} /var/qmail/control/envnoathost\n",
  restart_needed++;
  envnoathost_exists = 1;
  f07__does_recipient_contain_percent();
  restart_needed--;,
  "REAL_DOMAIN=${ENVNOATHOST_REAL_DOMAIN[${LEVEL_f03__does_envnoathost_exist}]}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} /var/qmail/control/envnoathost\n"
  "else\n"
  "  rm /var/qmail/control/envnoathost\n"
  "fi\n",
  "envnoathost does not exist",
  "if [ -f /var/qmail/control/envnoathost ]\n"
  "then\n"
  "  cp /var/qmail/control/envnoathost ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist}\n"
  "  rm /var/qmail/control/envnoathost\n"
  "fi\n",
  restart_needed++;
  envnoathost_exists = 0;
  f04__does_me_exist();
  restart_needed--;,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} /var/qmail/control/envnoathost\n"
  "fi\n")


DECISION(f02__does_recipient_have_domain_name,
  "recipient has domain name",
  "",
  if (relayclient_set)
    f07__does_recipient_contain_percent();
  else
    f05__is_domain_in_rcpthosts();,
  "",
  "recipient does not have a domain name",
  "REAL_DOMAIN=\"\"\n"
  "REAL_USERNAME=${REAL_USERNAME}${REAL_RELAYCLIENT}\n"
  "REAL_ADDRESS=${REAL_USERNAME}\n"
  "TO_DOMAIN=\"\"\n"
  "TO_ADDRESS=${TO_USERNAME}\n",
  if (spamdyke)
    reject_message(REJECTION_UNQUALIFIED);
  else
    f03__does_envnoathost_exist();,
  "")


DECISION(f01__is_relayclient_set,
  "RELAYCLIENT is set",
  "export RELAYCLIENT=\"foo\0\0\"\n"
  "REAL_RELAYCLIENT=${RELAYCLIENT}\n"
  "REAL_DOMAIN=${REAL_DOMAIN}${REAL_RELAYCLIENT}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n",
  relayclient_set = 1;
  f02__does_recipient_have_domain_name();
  relayclient_set = 0;,
  "export -n RELAYCLIENT\n",
  "RELAYCLIENT is not set",
  "export -n RELAYCLIENT\n",
  relayclient_set = 0;
  f02__does_recipient_have_domain_name();,
  "")

void usage()
  {
  fprintf(stderr,
    "USAGE: generator [ -v ] [ -b NUM ] [ -e NUM ] [ -p PREFIX ] [ -q ] [ -s ] target_dir\n"
    "\n"
    "Within target_dir, generator creates the test scripts necessary to completely\n"
    "test qmail's recipient validation procedure.  Two versions of each possible test\n"
    "can be created: one that tests the scenario with qmail only and one that tests\n"
    "the scenario with spamdyke.\n"
    "\n"
    "Generating the full list of tests will consume gigabytes of disk space.  Using\n"
    "the -b and -e options to limit the number of scripts is advisable.\n"
    "\n"
    "target_dir must be a directory and no test scripts must exist within it;\n"
    "generator will not overwrite any files.\n"
    "\n"
    "-b NUM\n"
    "  Start generating scripts at number NUM\n"
    "\n"
    "-e NUM\n"
    "  Stop generating scripts at number NUM\n"
    "\n"
    "-p PREFIX\n"
    "  Only generate scripts whose sequence of YNYN characters starts with PREFIX\n"
    "\n"
    "-q\n"
    "  Create the scripts to test with qmail only\n"
    "\n"
    "-s\n"
    "  Create the scripts to test with spamdyke\n"
    "\n"
    "-v\n"
    "  Produce verbose output during processing (may be given multiple times)\n"
    "\n"
    "Either -q or -s or both must be given.\n"
    );

  exit(1);

  return;
  }

void reinit()
  {
  strlen_comments = 0;
  comments[0] = '\0';
  strlen_pre = 0;
  pre[0] = '\0';
  strlen_post = 0;
  post[0] = '\0';
  strlen_path = 0;
  path[0] = '\0';
  test_name[0] = '\0';

  restart_needed = 0;
  relayclient_set = 0;
  using_defaultdelivery = 0;
  dash_emptyable = 1;
  dash_empty = 1;
  num_ext_hyphens = 0;
  num_username_hyphens = 0;
  forwards_only = 0;
  lines_in_file = 0;
  comments_in_file = 0;
  username_truncated = 0;

  return;
  }

int main(int argc, char *argv[])
  {
  int i;
  int do_qmail;
  int do_spamdyke;
  int opt;
  int tmp_int;

  opterr = 0;
  do_qmail = 0;
  do_spamdyke = 0;
  total_number = 0;
  next_number = 0;
  start_time = 0;
  max_strlen = 0;
  test_dir[0] = '\0';
  begin_num = -1;
  end_num = -1;
  strlen_target_prefix = 0;
  verbose = 0;

  while ((opt = getopt(argc, argv, "d:qsb:e:p:v")) != -1)
    {
    switch (opt)
      {
      case 'q':
        do_qmail = 1;
        break;
      case 's':
        do_spamdyke = 1;
        break;
      case 'b':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          begin_num = tmp_int;

        break;
      case 'e':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          end_num = tmp_int;

        break;
      case 'p':
        strlen_target_prefix = MINVAL(MAX_BUF, strlen(optarg));
        memcpy(target_prefix, optarg, strlen_target_prefix);
        target_prefix[strlen_target_prefix] = '\0';
        break;
      case 'v':
        verbose = 1;
        break;
      default:
        usage();
        break;
      }
    }

  if ((argv[optind] != NULL) &&
      (do_qmail ||
       do_spamdyke))
    {
    srandom(time(NULL));

    memcpy(test_dir, argv[optind], MINVAL(strlen(argv[optind]), MAX_BUF));
    for (i = 0; test_dir[i] != '\0'; i++);
    for (i--; test_dir[i] == '/'; i--)
      test_dir[i] = '\0';

    reinit();
    count_only = 1;

    if (do_qmail)
      {
      spamdyke = 0;
      f01__is_relayclient_set();
      }

    reinit();

    if (do_spamdyke)
      {
      spamdyke = 1;
      f01__is_relayclient_set();
      }

    total_number--;
    count_only = 0;

    if (MAX_BUF < max_strlen)
      {
      printf("ERROR: Largest buffer needed is %d bytes, allocation is set to %d\nIncrease MAX_BUF to fix this problem.\n", max_strlen, MAX_BUF);
      exit(0);
      }

    reinit();
    start_time = time(NULL);

    if (do_qmail)
      {
      spamdyke = 0;
      f01__is_relayclient_set();
      }

    reinit();

    if (do_spamdyke)
      {
      spamdyke = 1;
      f01__is_relayclient_set();
      }
    }
  else
    usage();

  return(0);
  }
