/*
  tofile -- a program to accept input from stdin and write to a file
  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/select.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#define MAX_BUF                         4095

#define STDIN_FD                        0
#define STDOUT_FD                       1

#define LOOP_SECS                       1

void usage()
  {
  printf(
    "USAGE: tofile [ -t SECS ] -f FILE\n"
    "\n"
    "-f FILE\n"
    "  Append all input from stdin to FILE\n"
    "-t SECS\n"
    "  Unconditionally exit after SECS seconds\n"
    );
  exit(0);
  }

int main(int argc, char *argv[])
  {
  int return_value;
  char *output_filename;
  FILE *output_file;
  char tmp_buf[MAX_BUF + 1];
  struct timeval tmp_timeval;
  fd_set read_fdset;
  int timeout_secs;
  long start_time;
  int tmp_int;
  int continue_looping;
  int read_result;
  int opt;

  return_value = 0;
  output_filename = NULL;
  output_file = NULL;
  timeout_secs = -1;
  continue_looping = 1;

  while ((opt = getopt(argc, argv, "f:t:")) != -1)
    {
    switch (opt)
      {
      case 'f':
        output_filename = optarg;
        break;
      case 't':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          timeout_secs = tmp_int;

        break;
      default:
        usage();
        break;
      }
    }

  if ((output_filename == NULL) ||
      (argv[optind] != NULL))
    usage();

  if ((output_file = fopen(output_filename, "a")) == NULL)
    {
    fprintf(stderr, "ERROR: unable to open %s: %s\n", output_filename, strerror(errno));
    return_value = 100;
    continue_looping = 0;
    }

  start_time = time(NULL);

  while (continue_looping)
    {
    tmp_timeval.tv_sec = LOOP_SECS;
    tmp_timeval.tv_usec = 0;

    FD_ZERO(&read_fdset);
    FD_SET(STDIN_FD, &read_fdset);

    if (select(STDIN_FD + 1, &read_fdset, NULL, NULL, &tmp_timeval) > 0)
      {
      if ((read_result = read(STDIN_FD, tmp_buf, MAX_BUF)) > 0)
        {
        tmp_buf[read_result] = '\0';
        fprintf(output_file, "%.*s", read_result, tmp_buf);
        }
      else if (read_result == 0)
        break;
      else
        {
        return_value = 100;
        break;
        }
      }

    if ((timeout_secs > 0) &&
        ((time(NULL) - start_time) > timeout_secs))
      continue_looping = 0;
    }

  if (output_file != NULL)
    {
    fclose(output_file);
    output_file = NULL;
    }

  return(return_value);
  }
